import unittest as ut
from torch_ga.cayley import blades_from_bases, get_cayley_tensor

dual_metric = [0]
dual_bases = ["0"]
dual_blades = ["", "0"]
dual_blade_degrees = [len(blade) for blade in dual_blades]


class TestDualCayleyTensor(ut.TestCase):
    def test_dual_blades_from_bases(self):
        blades, blade_degrees = blades_from_bases(dual_bases)
        self.assertCountEqual(blades, dual_blades)
        for blade, blade_degree in zip(dual_blades, dual_blade_degrees):
            blade_index = blades.index(blade)
            self.assertEqual(blade_degrees[blade_index], blade_degree)

    def test_cayley_tensor_correct(self):
        cayley, cayley_inner, cayley_outer = get_cayley_tensor(
            dual_metric, dual_bases, dual_blades)

        self.assertSequenceEqual(cayley.shape, [2, 2, 2])

        # Scalar * Scalar -> Scalar
        self.assertEqual(cayley[0, 0, 0], 1)
        self.assertEqual(cayley[0, 0, 1], 0)

        # Scalar * Dual -> Dual
        self.assertEqual(cayley[0, 1, 0], 0)
        self.assertEqual(cayley[0, 1, 1], 1)

        # Dual * Scalar -> Dual
        self.assertEqual(cayley[1, 0, 0], 0)
        self.assertEqual(cayley[1, 0, 1], 1)

        # Dual * Dual -> Zero
        self.assertEqual(cayley[1, 1, 0], 0)
        self.assertEqual(cayley[1, 1, 1], 0)




if __name__ == '__main__':
    ut.main()        


# from tfga import GeometricAlgebra
from torch_ga import GeometricAlgebra
import unittest as ut
import torch


pga_signature = [0, 1, 1, 1]

from torch.testing import assert_close
class TestDualGeometricAlgebraMultiply(ut.TestCase):
    def assertTensorsApproxEqual(self, a, b, tolerance=1e-4):
        print(f"assertTensorsApproxEqual(a={a},b={b})")
        assert_close(a,b,atol=tolerance, rtol=0), "%s not equal to %s" % (a, b)

    # def assertTensorsApproxEqual(self, a, b, tolerance=1e-4):
    #     self.assertTrue(tf.reduce_all(tf.abs(a - b) < tolerance),
    #                     "%s not equal to %s" % (a, b))

    def test_exp_eq_approx_exp_e01_e02(self):
        pga = GeometricAlgebra(pga_signature)

        # a = 3e01 + 5e02
        a = 3 * pga.e01 + 5 * pga.e02
        print("test_exp_eq_approx_exp_e01_e02")
        print(f"a={a}")
        # exp(a) = 1 + 3e01 + 5e02
        self.assertTensorsApproxEqual(pga.approx_exp(a), pga.exp(a))

    def test_exp_eq_approx_exp_e12_e23(self):
        pga = GeometricAlgebra(pga_signature)

        # a = 3e12 + 5e23
        a = 3 * pga.e12 + 5 * pga.e23
        print("test_exp_eq_approx_exp_e12_e23")
        print(f"a={a}")
        # exp(a) ~= 0.90 - 0.22e12 -0.37e23
        self.assertTensorsApproxEqual(pga.approx_exp(a), pga.exp(a))

    def test_inverse(self):
        pga = GeometricAlgebra(pga_signature)

        # a = 3e12 + 5e23
        a = 3 * pga.e12 + 5 * pga.e23

        # a_inv: -0.09*e_12 + -0.15*e_23
        a_inv = pga.inverse(a)

        # print("test_inverse")
        # print(f"a={a}")
        # print(f"a_inv={a_inv}")

        # a a_inv should be 1
        self.assertTensorsApproxEqual(pga.geom_prod(a, a_inv), 1 * pga.e(""))


if __name__ == '__main__':
    ut.main()    


import unittest as ut
import numpy as np
from torch_ga.cayley import blades_from_bases, get_cayley_tensor


sta_metric = [1, -1, -1, -1]
sta_bases = ["0", "1", "2", "3"]
sta_blades = [
    "",
    "0", "1", "2", "3",
    "01", "02", "03", "12", "13", "23",
    "012", "013", "023", "123",
    "0123"
]
sta_blade_degrees = [len(blade) for blade in sta_blades]


class TestStaCayleyTensor(ut.TestCase):
    def test_sta_blades_from_bases(self):
        blades, blade_degrees = blades_from_bases(sta_bases)
        self.assertCountEqual(blades, sta_blades)
        for blade, blade_degree in zip(sta_blades, sta_blade_degrees):
            blade_index = blades.index(blade)
            self.assertEqual(blade_degrees[blade_index], blade_degree)

    def test_sta_cayley_tensor_scalar_scalar_scalar(self):
        cayley, cayley_inner, cayley_outer = get_cayley_tensor(
            sta_metric, sta_bases, sta_blades)

        # Scalar * Scalar -> Scalar
        self.assertEqual(cayley[0, 0, 0], 1)
        self.assertTrue(np.all(cayley[0, 0, 1:] == 0))

    def test_sta_cayley_tensor_e12_e23_me13(self):
        cayley, cayley_inner, cayley_outer = get_cayley_tensor(
            sta_metric, sta_bases, sta_blades)

        # e12 * e23 -> -e13
        e12_index = sta_blades.index("12")
        e23_index = sta_blades.index("23")
        e13_index = sta_blades.index("13")
        self.assertEqual(cayley[e12_index, e23_index, e13_index], -1)
        self.assertTrue(np.all(cayley[e12_index, e23_index, :e13_index] == 0))
        self.assertTrue(
            np.all(cayley[e12_index, e23_index, e13_index+1:] == 0))

    def test_sta_cayley_tensor_e01_e23_e0123(self):
        cayley, cayley_inner, cayley_outer = get_cayley_tensor(
            sta_metric, sta_bases, sta_blades)

        # e01 * e23 -> e0123
        e01_index = sta_blades.index("01")
        e23_index = sta_blades.index("23")
        e0123_index = sta_blades.index("0123")
        self.assertEqual(cayley[e01_index, e23_index, e0123_index], 1)
        self.assertTrue(
            np.all(cayley[e01_index, e23_index, :e0123_index] == 0))
        self.assertTrue(
            np.all(cayley[e01_index, e23_index, e0123_index+1:] == 0))


if __name__ == '__main__':
    ut.main()        


"""Operations for constructing the cayley 3-tensor needed
for the geometric product. Used internally.
"""
from itertools import combinations
import numpy as np

from .blades import get_normal_ordered


def _collapse_same(x):
    for i in range(len(x) - 1):
        a, b = x[i], x[i + 1]
        if a == b:
            return False, x[:i] + x[i+2:], a
    return True, x, None


def _reduce_bases(a, b, metric):
    if a == "":
        return 1, b
    elif b == "":
        return 1, a

    combined = list(a + b)

    # Bring into normal order:
    sign, combined = get_normal_ordered(combined)

    done = False
    while not done:
        done, combined, combined_elem = _collapse_same(combined)
        if not done:
            sign *= metric[combined_elem]

    return sign, "".join(combined)


def blades_from_bases(vector_bases):
    all_combinations = [""]
    degrees = [0]
    for i in range(1, len(vector_bases) + 1):
        combs = combinations(vector_bases, i)
        combs = ["".join(c) for c in combs]
        all_combinations += combs
        degrees += [i] * len(combs)
    return all_combinations, degrees


def get_cayley_tensor(metric, bases, blades):
    num_blades = len(blades)

    t_geom = np.zeros((num_blades, num_blades, num_blades), dtype=np.int32)
    t_inner = np.zeros((num_blades, num_blades, num_blades), dtype=np.int32)
    t_outer = np.zeros((num_blades, num_blades, num_blades), dtype=np.int32)

    metric_dict = {v: metric[i] for i, v in enumerate(bases)}

    for a in blades:
        for b in blades:
            sign, result = _reduce_bases(a, b, metric_dict)
            a_index = blades.index(a)
            b_index = blades.index(b)
            out_index = blades.index(result)
            t_geom[a_index, b_index, out_index] = sign

            # Degree went down -> part of inner
            if len(result) == abs(len(a) - len(b)):
                t_inner[a_index, b_index, out_index] = sign

            # Degree went up -> part of outer
            if len(result) == len(a) + len(b):
                t_outer[a_index, b_index, out_index] = sign

    return t_geom, t_inner, t_outer



import unittest as ut
from io import BytesIO
from torch_ga.layers import (
    GeometricProductDense, GeometricSandwichProductDense,
    GeometricProductElementwise, GeometricSandwichProductElementwise,
    GeometricProductConv1D,
    GeometricAlgebraExp,
    GeometricToTensor, GeometricToTensorWithKind,
    TensorToGeometric, TensorWithKindToGeometric,
)
from torch_ga.blades import BladeKind
from torch_ga import GeometricAlgebra

import h5py
import torch
import torch.nn as nn
import torch.nn.functional as F

import torch
torch.manual_seed(0)



class TestKerasLayers(ut.TestCase):
    def assertTensorsEqual(self, a, b):
        # self.assertTrue(tf.reduce_all(a == b), "%s not equal to %s" % (a, b))
        print(f"assertTensorsEqual(a={a},b={b})")
        assert torch.all(a.squeeze() == b.squeeze()), "%s not equal to %s" % (a, b)

    def test_tensor_to_geometric(self):
        sta = GeometricAlgebra([1, -1, -1, -1])
        tensor = torch.ones([32, 4])
        gt_geom_tensor = torch.concat(
            [torch.zeros([32, 1]), torch.ones([32, 4]), torch.zeros([32, 11])],
            axis=-1
        )

        vector_blade_indices = [1, 2, 3, 4]

        tensor_to_geom_layer = TensorToGeometric(sta, vector_blade_indices)

        self.assertTensorsEqual(tensor_to_geom_layer(tensor), gt_geom_tensor)

    def test_tensor_with_kind_to_geometric(self):
        sta = GeometricAlgebra([1, -1, -1, -1])
        tensor = torch.ones([32, 4])
        gt_geom_tensor = torch.concat(
            [torch.zeros([32, 1]), torch.ones([32, 4]), torch.zeros([32, 11])],
            axis=-1
        )

        vector_blade_indices = [1, 2, 3, 4]

        tensor_kind_to_geom_layer = TensorWithKindToGeometric(
            sta, BladeKind.VECTOR)

        self.assertTensorsEqual(
            tensor_kind_to_geom_layer(tensor), gt_geom_tensor)

    def test_geometric_to_tensor(self):
        sta = GeometricAlgebra([1, -1, -1, -1])
        gt_tensor = torch.ones([32, 4])
        geom_tensor = torch.concat(
            [torch.zeros([32, 1]), torch.ones([32, 4]), torch.zeros([32, 11])],
            axis=-1
        )

        vector_blade_indices = [1, 2, 3, 4]

        geom_to_tensor_layer = GeometricToTensor(sta, vector_blade_indices)

        self.assertTensorsEqual(geom_to_tensor_layer(geom_tensor), gt_tensor)

    def test_geometric_to_tensor_with_kind(self):
        sta = GeometricAlgebra([1, -1, -1, -1])
        gt_tensor = torch.ones([32, 4])
        geom_tensor = torch.concat(
            [torch.zeros([32, 1]), torch.ones([32, 4]), torch.zeros([32, 11])],
            axis=-1
        )

        vector_blade_indices = [1, 2, 3, 4]

        geom_to_tensor_kind_layer = GeometricToTensorWithKind(
            sta, BladeKind.VECTOR)

        self.assertTensorsEqual(
            geom_to_tensor_kind_layer(geom_tensor), gt_tensor)

    def test_geometric_product_dense_v_v(self):
        sta = GeometricAlgebra([1, -1, -1, -1])

        geom_tensor = torch.concat(
            [torch.zeros([32, 6, 1]), torch.ones([32, 6, 4]), torch.zeros([32, 6, 11])],
            axis=-1
        )

        vector_blade_indices = [1, 2, 3, 4]

        geom_prod_layer = GeometricProductDense(
            sta, 8,
            blade_indices_kernel=vector_blade_indices,
            blade_indices_bias=vector_blade_indices,
            # bias_initializer=tf.keras.initializers.RandomNormal()
        )

        result = geom_prod_layer(geom_tensor)

        print(f"test_geometric_product_dense_v_v:")
        print(f"geom_tensor={geom_tensor}")
        print(f"result={result}")   

        # vector * vector + vector -> scalar + bivector + vector
        expected_result_indices = torch.tensor([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

        self.assertTrue(torch.all(sta.is_pure(result, expected_result_indices)))

    def test_geometric_product_dense_s_mv(self):
        sta = GeometricAlgebra([1, -1, -1, -1])

        geom_tensor = torch.concat(
            [torch.ones([20, 6, 1]), torch.zeros([20, 6, 15])],
            axis=-1
        )

        mv_blade_indices = list(range(16))

        geom_prod_layer = GeometricProductDense(
            sta, 8,
            blade_indices_kernel=mv_blade_indices,
            blade_indices_bias=mv_blade_indices
        )
        geom_prod_layer.build(geom_tensor.shape)

        result = geom_prod_layer(geom_tensor)

        print(f"test_geometric_product_dense_s_mv:")
        print(f"geom_tensor={geom_tensor}")
        print(f"result={result}")        


        # scalar * multivector + multivector -> multivector
        # Check that nothing is zero (it would be extremely unlikely
        # but not impossible to randomly get a zero here).
        assert torch.all(result != 0.0)
        # self.assertTrue(tf.reduce_all(result != 0.0))

    def test_geometric_product_dense_sequence(self):
        sta = GeometricAlgebra([1, -1, -1, -1])

        tensor = torch.ones([20, 6, 4])

        vector_blade_indices = [1, 2, 3, 4]
        mv_blade_indices = list(range(16))

        # vector * vector + vector -> scalar + bivector + vector
        scalar_bivector_blade_indices = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        sequence = nn.Sequential(*[
            TensorToGeometric(sta, blade_indices=vector_blade_indices),
            GeometricProductDense(
                sta, 8,
                blade_indices_kernel=vector_blade_indices,
                blade_indices_bias=vector_blade_indices,
                # bias_initializer=tf.keras.initializers.RandomNormal()
            ),
            GeometricToTensor(sta, blade_indices=scalar_bivector_blade_indices)
        ])
        for e in sequence:
            e.build(tensor.shape)

        result = sequence(tensor)

        print(f"test_geometric_product_dense_sequence:")
        print(f"tensor={tensor}")
        print(f"result={result}")        

        self.assertEqual(result.shape[-1], len(scalar_bivector_blade_indices))

    def test_geometric_sandwich_product_dense_v_v(self):
        sta = GeometricAlgebra([1, -1, -1, -1])

        geom_tensor = torch.concat(
            [torch.zeros([32, 6, 1]), torch.ones([32, 6, 4]), torch.zeros([32, 6, 11])],
            axis=-1
        )

        vector_blade_indices = [1, 2, 3, 4]

        result_indices = torch.concat([
            sta.get_kind_blade_indices(BladeKind.VECTOR),
            sta.get_kind_blade_indices(BladeKind.TRIVECTOR)
        ], axis=0)

        geom_prod_layer = GeometricSandwichProductDense(
            sta, 8,
            blade_indices_kernel=vector_blade_indices,
            blade_indices_bias=result_indices,
            # bias_initializer=tf.keras.initializers.RandomNormal()
        )

        result = geom_prod_layer(geom_tensor)

        print(f"test_geometric_sandwich_product_dense_v_v:")
        print(f"geom_tensor={geom_tensor}")
        print(f"result={result}")

        # vector * vector * ~vector + vector -> vector + trivector

        self.assertTrue(torch.all(sta.is_pure(result, result_indices)))


class TestKerasLayersSerializable(ut.TestCase):
    def assertTensorsEqual(self, a, b):
        # self.assertTrue(tf.reduce_all(a == b), "%s not equal to %s" % (a, b))
        assert torch.all(a == b), "%s not equal to %s" % (a, b)

    def _test_layer_serializable(self, layer, inputs):
        # Create algebra
        algebra = layer.algebra

        # Create model
        model = nn.Sequential(*[layer])

        # Predict on inputs to compare later
        layer.build(inputs.shape)
        model_output = model(inputs)

        # Serialize model to virtual file
        # model_file = h5py.File(BytesIO(), mode="w")
        # model.save(model_file)

        model_file = "./test_model.ph"
        torch.save(model.state_dict(), model_file)

        # Load model from stream
        # loaded_model = tf.keras.models.load_model(model_file)

        device = torch.device('cpu')
        loaded_model = nn.Sequential(*[layer])
        loaded_model.load_state_dict(torch.load(model_file, map_location=device))
        

        # Predict on same inputs as before
        loaded_output = loaded_model(inputs)

        # Check same output for original and loaded model
        self.assertTensorsEqual(model_output, loaded_output)

        # Check same recreated algebra
        self.assertTensorsEqual(
            # algebra.metric, loaded_model.layers[0].algebra.metric
            algebra.metric, loaded_model[0].algebra.metric
            )
        self.assertTensorsEqual(
            # algebra.cayley, loaded_model.layers[0].algebra.cayley
            algebra.cayley, loaded_model[0].algebra.cayley
            )

    def test_geom_dense_serializable(self):
        # Create algebra
        sta = GeometricAlgebra([1, -1, -1, -1])
        vector_blade_indices = [1, 2, 3, 4]
        mv_blade_indices = list(range(16))

        torch.manual_seed(0)
        # Create model
        self._test_layer_serializable(GeometricProductDense(
            sta, units=8,
            blade_indices_kernel=mv_blade_indices,
            blade_indices_bias=vector_blade_indices
        ), torch.randn(*[3, 6, sta.num_blades]))

    def test_sandwich_dense_serializable(self):
        # Create algebra
        sta = GeometricAlgebra([1, -1, -1, -1])
        vector_blade_indices = [1, 2, 3, 4]
        mv_blade_indices = list(range(16))

        # Create model
        torch.manual_seed(0)
        self._test_layer_serializable(GeometricSandwichProductDense(
            sta, units=8,
            blade_indices_kernel=mv_blade_indices,
            blade_indices_bias=vector_blade_indices
        ), torch.randn([3, 6, sta.num_blades]))

    def test_geom_elementwise_serializable(self):
        # Create algebra
        sta = GeometricAlgebra([1, -1, -1, -1])
        vector_blade_indices = [1, 2, 3, 4]
        mv_blade_indices = list(range(16))

        # Create model
        torch.manual_seed(0)

        self._test_layer_serializable(GeometricProductElementwise(
            sta,
            blade_indices_kernel=mv_blade_indices,
            blade_indices_bias=vector_blade_indices
        ), torch.randn(*[3, 6, sta.num_blades]))

    def test_sandwich_elementwise_serializable(self):
        # Create algebra
        sta = GeometricAlgebra([1, -1, -1, -1])
        vector_blade_indices = [1, 2, 3, 4]
        mv_blade_indices = list(range(16))

        # Create model
        torch.manual_seed(0)
        self._test_layer_serializable(GeometricSandwichProductElementwise(
            sta,
            blade_indices_kernel=mv_blade_indices,
            blade_indices_bias=vector_blade_indices
        ), torch.randn(*[3, 6, sta.num_blades]))

    def test_geom_prod_conv1d_serializable(self):
        # Create algebra
        sta = GeometricAlgebra([1, -1, -1, -1])
        vector_blade_indices = [1, 2, 3, 4]
        mv_blade_indices = list(range(16))

        # Create model
        torch.manual_seed(0)
        self._test_layer_serializable(GeometricProductConv1D(
            sta, filters=8, kernel_size=3,
            padding="same", stride=2,
            blade_indices_kernel=mv_blade_indices,
            blade_indices_bias=vector_blade_indices
        ), torch.randn(*[3, 8, 4, sta.num_blades]))

    def test_tensor_to_geom_serializable(self):
        # Create algebra
        sta = GeometricAlgebra([1, -1, -1, -1])
        vector_blade_indices = [1, 2, 3, 4]

        torch.manual_seed(0)
        # Create model
        self._test_layer_serializable(TensorToGeometric(
            sta, blade_indices=vector_blade_indices
        ), torch.randn(*[1, 2, 3, len(vector_blade_indices)]))

    def test_geom_to_tensor_serializable(self):
        # Create algebra
        sta = GeometricAlgebra([1, -1, -1, -1])
        vector_blade_indices = [1, 2, 3, 4]

        # Create model
        self._test_layer_serializable(GeometricToTensor(
            sta, blade_indices=vector_blade_indices
        ), torch.randn([1, 2, 3, sta.num_blades]))

    def test_geom_exp_serializable(self):
        # Create algebra
        ga = GeometricAlgebra([1, 1, 1])
        torch.manual_seed(0)

        inputs = ga.from_tensor_with_kind(
            torch.randn(*[3]), BladeKind.BIVECTOR
        )

        # Create model
        self._test_layer_serializable(GeometricAlgebraExp(
            ga
        ), inputs)



if __name__ == '__main__':
    ut.main()     


"""Defines the `MultiVector` class which is used as a convenience wrapper
for `GeometricAlgebra` operations.
"""
from __future__ import annotations
from typing import Union

from .blades import BladeKind


class MultiVector:
    """Wrapper for geometric algebra tensors using `GeometricAlgebra`
    operations in a less verbose way using operators.
    """

    def __init__(self, blade_values: tf.Tensor, algebra: GeometricAlgebra):
        """Initializes a MultiVector from a geometric algebra `tf.Tensor`
        and its corresponding `GeometricAlgebra`.

        Args:
            blade_values: Geometric algebra `tf.Tensor` with as many elements
            on its last axis as blades in the algebra
            algebra: `GeometricAlgebra` instance corresponding to the geometric
            algebra tensor
        """

        self._blade_values = blade_values
        self._algebra = algebra

    @property
    def tensor(self):
        """Geometric algebra tensor holding the values of this multivector."""
        return self._blade_values

    @property
    def algebra(self):
        """`GeometricAlgebra` instance this multivector belongs to."""
        return self._algebra

    @property
    def batch_shape(self):
        """Batch shape of the multivector (ie. the shape of all axes except
        for the last one in the geometric algebra tensor).
        """
        return self._blade_values.shape[:-1]

    def __len__(self) -> int:
        """Number of elements on the first axis of the geometric algebra
        tensor."""
        return self._blade_values.shape[0]

    def __iter__(self):
        for n in range(self._blade_values.shape[0]):
            # If we only have one axis left, return the
            # actual numbers, otherwise return a new
            # multivector.
            if self._blade_values.shape.ndims == 1:
                yield self._blade_values[n]
            else:
                yield MultiVector(
                    self._blade_values[n],
                    self._algebra
                )

    def __xor__(self, other: self) -> self:
        """Exterior product. See `GeometricAlgebra.ext_prod()`"""
        assert isinstance(other, MultiVector)

        return MultiVector(
            self._algebra.ext_prod(self._blade_values, other._blade_values),
            self._algebra
        )

    def __or__(self, other: self) -> self:
        """Inner product. See `GeometricAlgebra.inner_prod()`"""
        assert isinstance(other, MultiVector)

        return MultiVector(
            self._algebra.inner_prod(self._blade_values, other._blade_values),
            self._algebra
        )

    def __mul__(self, other: self) -> self:
        """Geometric product. See `GeometricAlgebra.geom_prod()`"""
        assert isinstance(other, MultiVector)

        return MultiVector(
            self._algebra.geom_prod(self._blade_values, other._blade_values),
            self._algebra
        )

    def __truediv__(self, other: self) -> self:
        """Division, ie. multiplication with the inverse."""
        assert isinstance(other, MultiVector)

        return MultiVector(
            self._algebra.geom_prod(
                self._blade_values,
                self._algebra.inverse(other._blade_values)
            ),
            self._algebra
        )

    def __and__(self, other: self) -> self:
        """Regressive product. See `GeometricAlgebra.reg_prod()`"""
        assert isinstance(other, MultiVector)

        return MultiVector(
            self._algebra.reg_prod(self._blade_values, other._blade_values),
            self._algebra
        )

    def __invert__(self) -> self:
        """Reversion. See `GeometricAlgebra.reversion()`"""
        return MultiVector(
            self._algebra.reversion(self._blade_values),
            self._algebra
        )

    def __neg__(self) -> self:
        """Negation."""
        return MultiVector(
            -self._blade_values,
            self._algebra
        )

    def __add__(self, other: self) -> self:
        """Addition of multivectors."""
        assert isinstance(other, MultiVector)

        return MultiVector(
            self._blade_values + other._blade_values,
            self._algebra
        )

    def __sub__(self, other: self) -> self:
        """Subtraction of multivectors."""
        assert isinstance(other, MultiVector)

        return MultiVector(
            self._blade_values - other._blade_values,
            self._algebra
        )

    def __pow__(self, n: int) -> self:
        """Multivector raised to an integer power."""
        return MultiVector(
            self._algebra.int_pow(self._blade_values, n),
            self._algebra
        )

    def __getitem__(self, key: Union[str, List[str]]) -> self:
        """`MultiVector` with only passed blade names as non-zeros."""
        return MultiVector(
            self._algebra.keep_blades_with_name(self._blade_values, key),
            self._algebra
        )

    def __call__(self, key: Union[str, List[str]]):
        """`tf.Tensor` with passed blade names on last axis."""
        return self._algebra.select_blades_with_name(self._blade_values, key)

    def __repr__(self) -> str:
        return self._algebra.mv_repr(self._blade_values)

    def inverse(self) -> self:
        """Inverse. See `GeometricAlgebra.inverse()`."""
        return MultiVector(
            self._algebra.inverse(self._blade_values),
            self._algebra
        )

    def simple_inverse(self) -> self:
        """Simple inverse. See `GeometricAlgebra.simple_inverse()`."""
        return MultiVector(
            self._algebra.simple_inverse(self._blade_values),
            self._algebra
        )

    def dual(self) -> self:
        """Dual. See `GeometricAlgebra.dual()`."""
        return MultiVector(
            self._algebra.dual(self._blade_values),
            self._algebra
        )

    def conjugation(self) -> self:
        """Conjugation. See `GeometricAlgebra.conjugation()`."""
        return MultiVector(
            self._algebra.conjugation(self._blade_values),
            self._algebra
        )

    def grade_automorphism(self) -> self:
        """Grade automorphism. See `GeometricAlgebra.grade_automorphism()`."""
        return MultiVector(
            self._algebra.grade_automorphism(self._blade_values),
            self._algebra
        )

    def approx_exp(self, order: int = 50) -> self:
        """Approximate exponential. See `GeometricAlgebra.approx_exp()`."""
        return MultiVector(
            self._algebra.approx_exp(self._blade_values, order=order),
            self._algebra
        )

    def exp(self, square_scalar_tolerance: Union[float, None] = 1e-4) -> self:
        """Exponential. See `GeometricAlgebra.exp()`."""
        return MultiVector(
            self._algebra.exp(
                self._blade_values,
                square_scalar_tolerance=square_scalar_tolerance
            ),
            self._algebra
        )

    def approx_log(self, order: int = 50) -> self:
        """Approximate logarithm. See `GeometricAlgebra.approx_log()`."""
        return MultiVector(
            self._algebra.approx_log(self._blade_values, order=order),
            self._algebra
        )

    def is_pure_kind(self, kind: BladeKind) -> bool:
        """Whether the `MultiVector` is of a pure kind."""
        return self._algebra.is_pure_kind(self._blade_values, kind=kind)

    def geom_conv1d(self, kernel: self,
                    stride: int, padding: str,
                    dilations: Union[int, None] = None) -> self:
        """1D convolution. See `GeometricAlgebra.geom_conv1d().`"""
        return MultiVector(
            self._algebra.geom_conv1d(
                self._blade_values, kernel._blade_values,
                stride=stride, padding=padding, dilations=dilations
            ),
            self._algebra
        )



"""Blade-related definitions and functions used across the library."""
from enum import Enum
from typing import List, Tuple, Union
import torch


class BladeKind(Enum):
    """Kind of blade depending on its degree."""
    MV = "mv"
    EVEN = "even"
    ODD = "odd"
    SCALAR = "scalar"
    VECTOR = "vector"
    BIVECTOR = "bivector"
    TRIVECTOR = "trivector"
    PSEUDOSCALAR = "pseudoscalar"
    PSEUDOVECTOR = "pseudovector"
    PSEUDOBIVECTOR = "pseudobivector"
    PSEUDOTRIVECTOR = "pseudotrivector"


def get_blade_repr(blade_name: str) -> str:
    """Returns the representation to use
    for a given blade.

    Examples:
    - `"12"` -> `"e_12"`
    - `""` -> `"1"`

    Args:
        blade_name: name of the blade in the algebra (eg. `"12"`)

    Returns:
        Representation to use for a given blade
    """
    if blade_name == "":
        return "1"
    return "e_%s" % blade_name


def is_blade_kind(blade_degrees: torch.Tensor, kind: List[Union[BladeKind, str]], max_degree: int) -> torch.Tensor:
    """Finds a boolean mask for whether blade degrees are of a given kind.

    Args:
        blade_degrees: list of blade degrees
        kind: kind of blade to check for
        max_degree: maximum blade degree in the algebra

    Returns:
        boolean mask for whether blade degrees are of a given kind
    """
    # Convert kind to string representation
    # for comparison.
    kind = kind.value if isinstance(kind, BladeKind) else kind

    if kind == BladeKind.MV.value:
        # return torch.constant(True, shape=[len(blade_degrees)])
        return torch.full([len(blade_degrees)],True)
    elif kind == BladeKind.EVEN.value:
        return blade_degrees % 2 == 0
    elif kind == BladeKind.ODD.value:
        return blade_degrees % 2 == 1
    elif kind == BladeKind.SCALAR.value:
        return blade_degrees == 0
    elif kind == BladeKind.VECTOR.value:
        return blade_degrees == 1
    elif kind == BladeKind.BIVECTOR.value:
        return blade_degrees == 2
    elif kind == BladeKind.TRIVECTOR.value:
        return blade_degrees == 3
    elif kind == BladeKind.PSEUDOSCALAR.value:
        return blade_degrees == max_degree
    elif kind == BladeKind.PSEUDOVECTOR.value:
        return blade_degrees == max_degree - 1
    elif kind == BladeKind.PSEUDOBIVECTOR.value:
        return blade_degrees == max_degree - 2
    elif kind == BladeKind.PSEUDOTRIVECTOR.value:
        return blade_degrees == max_degree - 3
    raise Exception("Unknown blade kind: %s" % kind)


def set_diff(a,b):
    combined = torch.cat((a.squeeze(), b.squeeze()))
    uniques, counts = combined.unique(return_counts=True)
    difference = uniques[counts == 1]
    # intersection = uniques[counts > 1]    
    return difference


def invert_blade_indices(num_blades: int, blade_indices: torch.Tensor) -> torch.Tensor:
    """Returns all blade indices except for the given ones.

    Args:
        num_blades: Total number of blades in the algebra
        blade_indices: blade indices to exclude

    Returns:
        All blade indices except for the given ones
    """

    # all_blades = tf.range(num_blades, dtype=blade_indices.dtype)
    # return tf.sparse.to_dense(tf.sets.difference(
    #     tf.expand_dims(all_blades, axis=0),
    #     tf.expand_dims(blade_indices, axis=0)
    # ))[0]

    all_blades = torch.arange(num_blades, dtype=blade_indices.dtype)
    return set_diff(all_blades.unsqueeze(0), blade_indices.unsqueeze(0))[0]

# pip install functorch
# from functorch import vmap

def get_blade_of_kind_indices(blade_degrees: torch.Tensor, kind: BladeKind,
                              max_degree: int, invert: bool = False) -> torch.Tensor:
    """Finds a boolean mask for whether blades are of a given kind.

    Args:
        blade_degrees: List of blade degrees
        kind: kind of blade for which the mask will be true
        max_degree: maximum blade degree in the algebra
        invert: whether to invert the result

    Returns:
        boolean mask for whether blades are of a given kind
    """
    # cond = is_blade_kind(blade_degrees, kind, max_degree)
    # cond = tf.math.logical_xor(cond, invert)
    # return tf.where(cond)[:, 0]
    
    # cond = is_blade_kind(blade_degrees, kind, max_degree)
    # cond = torch.math.logical_xor(cond, invert)
    # return torch.where(cond)[:, 0]

    # cond = torch.vmap(is_blade_kind(blade_degrees, kind, max_degree))
    # cond = is_blade_kind(blade_degrees, kind, max_degree))
    # cond = cond(invert)
    # return torch.where(cond)[:, 0]

    # print(blade_degrees.shape)
    if False: print("get_blade_of_kind_indices:blade_degrees:",blade_degrees,"kind:",kind)
    cond = is_blade_kind(blade_degrees, kind, max_degree)
    # print("cond:",cond)
    # print(f"cond.shape={cond.shape}")
    cond = torch.logical_xor(cond,invert*torch.ones_like(cond))
    # print(f"cond.shape={cond.shape}")
    # print(f"cond.nonzero().shape={cond.nonzero().shape}")
    # print("cond:",cond)
    # print(cond.shape)
    # return torch.where(cond)[:, 0]
    # return cond[:, 0]
    return cond.nonzero().squeeze()
    # return cond


def _normal_swap(x: List[str]) -> List[str]:
    """Swaps the first unordered blade pair and returns the new list as well
    as whether a swap was performed."""
    for i in range(len(x) - 1):
        a, b = x[i], x[i + 1]
        if a > b:  # string comparison
            x[i], x[i+1] = b, a
            return False, x
    return True, x


def get_normal_ordered(blade_name: str) -> Tuple[int, str]:
    """Returns the normal ordered blade name and its sign.
    Example: 21 => -1, 12

    Args:
        blade_name: Blade name for which to return normal ordered
        name and sign

    Returns:
        sign: sign of the blade
        blade_name: normalized name of the blade
    """
    blade_name = list(blade_name)
    sign = -1
    done = False
    while not done:
        sign *= -1
        done, blade_name = _normal_swap(blade_name)
    return sign, "".join(blade_name)


def get_blade_indices_from_names(blade_names: List[str],
                                 all_blade_names: List[str]) -> torch.Tensor:
    """Finds blade signs and indices for given blade names in a list of blade
    names. Blade names can be unnormalized and their correct sign will be
    returned.

    Args:
        blade_names: Blade names to return indices for. May be unnormalized.
        all_blade_names: Blade names to use as index

    Returns:
        blade_signs: signs for the passed blades in same order as passed
        blade_indices: blade indices in the same order as passed
    """
    signs_and_names = [get_normal_ordered(b) for b in blade_names]

    blade_signs = [sign for sign, blade_name in signs_and_names]

    blade_indices = [
        all_blade_names.index(blade_name) for sign, blade_name in signs_and_names
    ]

    return (torch.tensor(blade_signs, dtype=torch.float32),
            torch.tensor(blade_indices, dtype=torch.int64))





import unittest as ut
import torch
from torch_ga import GeometricAlgebra


dual_metric = [0]
dual_bases = ["0"]
dual_blades = ["", "0"]
dual_blade_degrees = [len(blade) for blade in dual_blades]

from torch.testing import assert_close

class TestDualGeometricAlgebraMultiply(ut.TestCase):
    def assertTensorsEqual(self, a, b):
        print(f"assertTensorsEqual(a={a},b={b})")
        # print(f"a={a},b={b}")
        assert_close(a,b)

        # self.assertTrue(
        #     # tf.reduce_all(a == b), 
        #     assert_close(a,b),
        #     "%s not equal to %s" % (a, b))

    def test_mul_mv_mv(self):
        ga = GeometricAlgebra(metric=dual_metric)

        zero = ga.from_scalar(0.0)
        one = ga.from_scalar(1.0)
        eps = ga.from_tensor_with_kind(torch.ones(1), kind="pseudoscalar").squeeze()
        ten = ga.from_scalar(10.0)

        self.assertTensorsEqual(ga.geom_prod(eps, eps), zero)
        self.assertTensorsEqual(ga.geom_prod(one, one), one)
        self.assertTensorsEqual(ga.geom_prod(zero, one), zero)
        self.assertTensorsEqual(ga.geom_prod(one, zero), zero)
        self.assertTensorsEqual(ga.geom_prod(one, eps), eps)
        self.assertTensorsEqual(ga.geom_prod(eps, one), eps)
        self.assertTensorsEqual(ga.geom_prod(zero, zero), zero)
        self.assertTensorsEqual(ga.geom_prod(ten, zero), zero)
        self.assertTensorsEqual(ga.geom_prod(zero, ten), zero)
        self.assertTensorsEqual(
            ga.geom_prod(ga.geom_prod(ten, eps), eps),
            zero
        )
        self.assertTensorsEqual(ga.geom_prod(ten, one), ten)
        self.assertTensorsEqual(ga.geom_prod(one, ten), ten)

    def test_mul_tf_mv(self):
        ga = GeometricAlgebra(metric=dual_metric)

        zero = ga.from_scalar(0.0)
        one = ga.from_scalar(1.0)

        # eps = ga.from_tensor_with_kind(torch.ones(1), kind="pseudoscalar")
        eps = ga.from_tensor_with_kind(torch.tensor([1]), kind="pseudoscalar").squeeze()
        ten = ga.from_scalar(10.0)
        if False:        
            print(f"one={one}")
            print(f"zero={zero}")
            print(f"eps={eps}")
            print(f"ten={ten}")

        zero_tf = torch.tensor([0, 0], dtype=torch.float32)
        one_tf = torch.tensor([1, 0], dtype=torch.float32)
        eps_tf = torch.tensor([0, 1], dtype=torch.float32)
        ten_tf = torch.tensor([10, 0], dtype=torch.float32)
        if False:        
            print(f"zero_tf={zero_tf}")
            print(f"one_tf={one_tf}")
            print(f"eps_tf={eps_tf}")
            print(f"ten_tf={ten_tf}")

            print(f"ga.geom_prod(one, one_tf)={ga.geom_prod(one, one_tf)}")

        self.assertTensorsEqual(ga.geom_prod(one, one_tf), one)
        self.assertTensorsEqual(ga.geom_prod(one_tf, one), one)
        self.assertTensorsEqual(ga.geom_prod(zero, one_tf), zero)
        self.assertTensorsEqual(ga.geom_prod(one_tf, zero), zero)
        self.assertTensorsEqual(ga.geom_prod(zero_tf, one), zero)
        self.assertTensorsEqual(ga.geom_prod(one, zero_tf), zero)
        self.assertTensorsEqual(ga.geom_prod(one_tf, eps), eps)
        self.assertTensorsEqual(ga.geom_prod(eps, one_tf), eps)
        self.assertTensorsEqual(ga.geom_prod(zero_tf, zero), zero)
        self.assertTensorsEqual(ga.geom_prod(zero, zero_tf), zero)
        self.assertTensorsEqual(ga.geom_prod(ten_tf, zero), zero)
        self.assertTensorsEqual(ga.geom_prod(zero, ten_tf), zero)
        self.assertTensorsEqual(ga.geom_prod(ten, zero_tf), zero)
        self.assertTensorsEqual(ga.geom_prod(zero_tf, ten), zero)
        self.assertTensorsEqual(
            ga.geom_prod(ga.geom_prod(ten_tf, eps), eps),
            zero
        )
        self.assertTensorsEqual(ga.geom_prod(ten_tf, one), ten)
        self.assertTensorsEqual(ga.geom_prod(one, ten_tf), ten)
        self.assertTensorsEqual(ga.geom_prod(ten, one_tf), ten)
        self.assertTensorsEqual(ga.geom_prod(one_tf, ten), ten)


class TestDualGeometricAlgebraMisc(ut.TestCase):
    def assertTensorsEqual(self, a, b):
        # self.assertTrue(tf.reduce_all(a == b), "%s not equal to %s" % (a, b))
        self.assertTrue(torch.all(a == b), "%s not equal to %s" % (a, b))

    def test_auto_diff_square(self):
        """Test automatic differentiation using
        dual numbers for the square function.
        f(x) = x^2
        f'(x) = d/dx f(x) = 2x
        """
        ga = GeometricAlgebra(metric=dual_metric)

        one = ga.from_scalar(1.0)
        five = ga.from_scalar(5.0)
        eps = ga.from_tensor_with_kind(torch.ones(1), kind="pseudoscalar")

        x = one + eps
        if False:        
            print(f"one={one}")
            print(f"eps={eps}")
            print(f"x = one + eps={x}")        

        # f(1) = 1^2 = 1, f'(1) = 2
        x_squared = ga.geom_prod(x, x)
        self.assertTensorsEqual(ga.select_blades_with_name(x_squared, ""), 1.0)
        self.assertTensorsEqual(ga.select_blades_with_name(x_squared, "0"), 2.0)

        y = five + eps

        # f(5) = 5^2 = 25, f'(5) = 10
        y_squared = ga.geom_prod(y, y)
        self.assertTensorsEqual(ga.select_blades_with_name(y_squared, ""), 25.0)
        self.assertTensorsEqual(ga.select_blades_with_name(y_squared, "0"), 10.0)

    def test_batched_auto_diff_square(self):
        """Test automatic differentiation using
        dual numbers for the square function.
        Use batch with identical elements.
        f(x) = x^2
        f'(x) = d/dx f(x) = 2x
        """
        ga = GeometricAlgebra(metric=dual_metric)

        one = ga.from_tensor_with_kind(torch.ones([3, 4, 1]), kind="scalar")
        five = ga.from_tensor_with_kind(torch.full([3, 4, 1], 5.0), kind="scalar")
        eps = ga.from_tensor_with_kind(torch.ones([3, 4, 1]), kind="pseudoscalar")
        if False:        
            print(f"five={five},one={one}")
            print(f"five.shape={five.shape},one.shape={one.shape}")

        x = one + eps
        if False:        
            print(f"one={one}")
            print(f"eps={eps}")
            print(f"x = one + eps={x}")

        # f(1) = 1^2 = 1, f'(1) = 2
        x_squared = ga.geom_prod(x, x)

        if False:        
            print(f"x_squared={x_squared}")
            print(f"x_squared.shape={x_squared.shape}")
            _a = ga.select_blades_with_name(x_squared, '')
            _b = ga.select_blades_with_name(x_squared, '0')
            print(f"ga.select_blades_with_name(x_squared, '')={_a}")
            print(f"_a.shape={_a.shape}")
            print(f"ga.select_blades_with_name(x_squared, '0')={_b}")
            print(f"_b.shape={_b.shape}")

        self.assertTensorsEqual(ga.select_blades_with_name(x_squared, ""), 1.0)
        self.assertTensorsEqual(ga.select_blades_with_name(x_squared, "0"), 2.0)

        y = five + eps

        # f(5) = 5^2 = 25, f'(5) = 10
        y_squared = ga.geom_prod(y, y)
        self.assertTensorsEqual(ga.select_blades_with_name(y_squared, ""), 25.0)
        self.assertTensorsEqual(ga.select_blades_with_name(y_squared, "0"), 10.0)

    def test_mul_inverse(self):
        ga = GeometricAlgebra(metric=dual_metric)

        # a = 2
        a = ga.from_tensor_with_kind(torch.full([1], 2.0), kind="scalar")
        if False: print(f"a={a},torch.full([1], 2.0)={torch.full([1], 2.0)}")


        # b = 3 + 3e0
        b = ga.from_tensor_with_kind(torch.full([2], 3.0), kind="mv")

        # a * b = 2 * (3 + 3e0) = 6 + 6e0
        c = ga.geom_prod(a, b)
        self.assertTensorsEqual(c, ga.from_scalar(6.0) + 6.0 * ga.e("0"))
        if False:
            print(f"a={a}")
            print(f"b={b}")
            print(f"c=a*b={c}")

        # a^-1 = 1 / 2
        a_inv = ga.inverse(a)
        self.assertTensorsEqual(ga.select_blades_with_name(a_inv, ""), 0.5)
        if False:
            print(f"a={a}")
            print(f"a_inv={a_inv}")

        # c = a * b
        # => a_inv * c = b
        self.assertTensorsEqual(ga.geom_prod(a_inv, c), b)
        if False:
            print(f"b={b}")
            print(f"a_inv * c={ga.geom_prod(a_inv, c)}")

        # Since a is scalar, should commute too.
        # => c * a_inv = b
        self.assertTensorsEqual(ga.geom_prod(c, a_inv), b)
        if False:
            print(f"b={b}")
            print(f"a_inv * c={ga.geom_prod(c, a_inv)}")

        # b is not simply invertible (because it does not square to a scalar)
        # and will throw an exception
        self.assertRaises(Exception, ga.simple_inverse, b)

        # b is invertible with the shirokov inverse
        if False: print(f"b={b}")
        b_inv = ga.inverse(b)
        if False: print(f"b_inv={b_inv}")
        self.assertTensorsEqual(ga.geom_prod(b, b_inv), 1 * ga.e(""))


if __name__ == '__main__':
    ut.main()        


"""
PyTorch Geometric Algebra

An python library for Geometric Algebra in Pytorch
"""

__version__ = "0.0.2"
__author__ = 'Francesco Alesiani'
__credits__ = 'TF-GA (https://doi.org/10.5281/zenodo.3902404)'

from .torch_ga import GeometricAlgebra


from setuptools import setup
import os

# Read README.md
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as readme_file:
    long_description = readme_file.read()

setup(
    name="torch_ga",
    packages=["torch_ga"],
    extras_require={
        "torch": ["pytorch>=1.12.1"],
    },
    description="Clifford and Geometric Algebra with PyTorch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="0.0.2",
    url="https://github.com/falesiani/torch_ga",
    author="Francesco Alesiani",
    author_email="francesco.alesiani@neclab.eu",
    license="MIT",
    keywords="geometric-algebra clifford-algebra pytorch multi-vector para-vector mathematics machine-learning",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",

        "Programming Language :: Python :: 3",

        "Intended Audience :: Education",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",

        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",

        "Topic :: Software Development",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)



"""Provides classes and operations for performing geometric algebra
with TensorFlow.

The `GeometricAlgebra` class is used to construct the algebra given a metric.
It exposes methods for operating on `torch.Tensor` instances where their last
axis is interpreted as blades of the algebra.
"""
from typing import List, Any, Union, Optional
import numbers
import numpy as np
import torch
# import einops

from .cayley import get_cayley_tensor, blades_from_bases
from .blades import (
    BladeKind, get_blade_of_kind_indices, get_blade_indices_from_names,
    get_blade_repr, invert_blade_indices
)
from .mv_ops import mv_multiply, mv_reversion, mv_grade_automorphism, mv_conv1d, f_mv_conv1d, mv_multiply_element_wise
from .mv import MultiVector


class GeometricAlgebra:
    """Class used for performing geometric algebra operations on `torch.Tensor` instances.
    Exposes methods for operating on `torch.Tensor` instances where their last
    axis is interpreted as blades of the algebra.
    Holds the metric and other quantities derived from it.
    """

    def __init__(self, metric: List[float]):
        """Creates a GeometricAlgebra object given a metric.
        The algebra will have as many basis vectors as there are
        elements in the metric.

        Args:
            metric: Metric as a list. Specifies what basis vectors square to
        """
        self._metric = torch.tensor(metric, dtype=torch.float32)

        self._num_bases = len(metric)
        self._bases = list(map(str, range(self._num_bases)))

        self._blades, self._blade_degrees = blades_from_bases(self._bases)
        self._blade_degrees = torch.tensor(self._blade_degrees)
        self._num_blades = len(self._blades)
        self._max_degree = self._blade_degrees.max()

        # [Blades, Blades, Blades]
        _list = get_cayley_tensor(self.metric, self._bases, self._blades)
        # print(_list)
        if type(_list) in [list,tuple]:
            _list = np.array(_list)
        self._cayley, self._cayley_inner, self._cayley_outer = torch.tensor(
            _list,
            dtype=torch.float32
        )

        self._blade_mvs = torch.eye(self._num_blades)
        self._basis_mvs = self._blade_mvs[1:1+self._num_bases]

        # Find the dual by looking at the anti-diagonal in the Cayley tensor.
        self._dual_blade_indices = []
        self._dual_blade_signs = []

        for blade_index in range(self._num_blades):
            dual_index = self.num_blades - blade_index - 1
            anti_diag = self._cayley[blade_index, dual_index]
            # dual_sign = tf.gather(anti_diag, tf.where(
            #     anti_diag != 0.0)[..., 0])[..., 0]
            dual_sign = anti_diag[torch.where(anti_diag != 0.0)]

            self._dual_blade_indices.append(dual_index)
            self._dual_blade_signs.append(dual_sign)

        self._dual_blade_indices = torch.tensor(
            self._dual_blade_indices, dtype=torch.int64)
        self._dual_blade_signs = torch.tensor(
            self._dual_blade_signs, dtype=torch.float32)

    def print(self, *args, **kwargs):
        """Same as the default `print` function but formats `torch.Tensor`
        instances that have as many elements on their last axis
        as the algebra has blades using `mv_repr()`.
        """
        def _is_mv(arg):
            return isinstance(arg, torch.Tensor) and len(arg.shape) > 0 and arg.shape[-1] == self.num_blades
        new_args = [self.mv_repr(arg) if _is_mv(arg) else arg for arg in args]

        print(*new_args, **kwargs)

    @property
    def metric(self) -> torch.Tensor:
        """Metric list which contains the number that each
        basis vector in the algebra squares to
        (ie. the diagonal of the metric tensor).
        """
        return self._metric

    @property
    def cayley(self) -> torch.Tensor:
        """`MxMxM` tensor where `M` is the number of basis
        blades in the algebra. Used for calculating the
        geometric product:

        `a_i, b_j, cayley_ijk -> c_k`
        """
        return self._cayley

    @property
    def cayley_inner(self) -> torch.Tensor:
        """Analagous to cayley but for inner product."""
        return self._cayley_inner

    @property
    def cayley_outer(self) -> torch.Tensor:
        """Analagous to cayley but for outer product."""
        return self._cayley_outer

    @property
    def blades(self) -> List[str]:
        """List of all blade names.

        Blades are all possible independent combinations of
        basis vectors. Basis vectors are named starting
        from `"0"` and counting up. The scalar blade is the
        empty string `""`.

        Example
        - Bases: `["0", "1", "2"]`
        - Blades: `["", "0", "1", "2", "01", "02", "12", "012"]`
        """
        return self._blades

    @property
    def blade_mvs(self) -> torch.Tensor:
        """List of all blade tensors in the algebra."""
        return self._blade_mvs

    @property
    def dual_blade_indices(self) -> torch.Tensor:
        """Indices of the dual blades for each blade."""
        return self._dual_blade_indices

    @property
    def dual_blade_signs(self) -> torch.Tensor:
        """Signs of the dual blades for each blade."""
        return self._dual_blade_signs

    @property
    def num_blades(self) -> int:
        """Total number of blades in the algebra."""
        return self._num_blades

    @property
    def blade_degrees(self) -> torch.Tensor:
        """List of blade-degree for each blade in the algebra."""
        return self._blade_degrees

    @property
    def max_degree(self) -> int:
        """Highest blade degree in the algebra."""
        return self._max_degree

    @property
    def basis_mvs(self) -> torch.Tensor:
        """List of basis vectors as torch.Tensor."""
        return self._basis_mvs

    def get_kind_blade_indices(self, kind: BladeKind, invert: bool = False) -> torch.Tensor:
        """Find all indices of blades of a given kind in the algebra.

        Args:
            kind: kind of blade to give indices for
            invert: whether to return all blades not of the kind

        Returns:
            indices of blades of a given kind in the algebra
        """
        return get_blade_of_kind_indices(self.blade_degrees, kind, self.max_degree, invert=invert)

    def get_blade_indices_of_degree(self, degree: int) -> torch.Tensor:
        """Find all indices of blades of the given degree.

        Args:
            degree: degree to return blades for

        Returns:
            indices of blades with the given degree in the algebra
        """
        # return tf.gather(tf.range(self.num_blades), tf.where(self.blade_degrees == degree)[..., 0])
        return torch.range(self.num_blades)[torch.where(self.blade_degrees == degree)[..., 0]]

    def is_pure(self, tensor: torch.Tensor, blade_indices: torch.Tensor) -> bool:
        """Returns whether the given tensor is purely of the given blades
        and has no non-zero values for blades not in the given blades.

        Args:
            tensor: tensor to check purity for
            blade_indices: blade indices to check purity for

        Returns:
            Whether the tensor is purely of the given blades
            and has no non-zero values for blades not in the given blades
        """
        # tensor = torch.tensor(tensor, dtype=torch.float32)
        tensor = tensor.to(dtype=torch.float32)
        if not type(blade_indices) in [torch.Tensor]:
            blade_indices = torch.tensor(blade_indices)
            
        blade_indices = blade_indices.to(dtype=torch.int64)

        # blade_indices = torch.tensor(
        #     blade_indices, dtype=torch.int64)

        inverted_blade_indices = invert_blade_indices(
            self.num_blades, blade_indices)

        # return tf.reduce_all(tf.gather(
        #     tensor,
        #     inverted_blade_indices,
        #     axis=-1
        # ) == 0)
        return (tensor[inverted_blade_indices]==0).sum(dim=-1)

    def is_pure_kind(self, tensor: torch.Tensor, kind: BladeKind) -> bool:
        """Returns whether the given tensor is purely of a given kind
        and has no non-zero values for blades not of the kind.

        Args:
            tensor: tensor to check purity for
            kind: kind of blade to check purity for

        Returns:
            Whether the tensor is purely of a given kind
            and has no non-zero values for blades not of the kind
        """
        # tensor = torch.tensor(tensor, dtype=torch.float32)
        tensor = tensor.to(dtype=torch.float32)
        inverted_kind_indices = self.get_kind_blade_indices(kind, invert=True)
        # print(f"tensor={tensor}")
        # print(f"kind={kind}")
        # print(f"inverted_kind_indices={inverted_kind_indices.T}")
        # print(f"inverted_kind_indices.shape={inverted_kind_indices.shape}")
        # print(f"tensor[inverted_kind_indices]={tensor[inverted_kind_indices].T}")
        # print(f"tensor[inverted_kind_indices].shape={tensor[inverted_kind_indices].shape}")
        # print(f"tensor[inverted_kind_indices]==0={tensor[inverted_kind_indices].T==0}")

        # return tf.reduce_all(tf.gather(
        #     tensor,
        #     inverted_kind_indices,
        #     axis=-1
        # ) == 0)
        return (tensor[inverted_kind_indices]==0).sum(dim=-1)

    # def from_tensor(self, tensor: torch.Tensor, blade_indices: torch.Tensor) -> torch.Tensor:
    #     """Creates a geometric algebra torch.Tensor from a torch.Tensor and blade
    #     indices. The blade indices have to align with the last axis of the
    #     tensor.

    #     Args:
    #         tensor: torch.Tensor to take as values for the geometric algebra tensor
    #         blade_indices: Blade indices corresponding to the tensor. Can
    #         be obtained from blade names eg. using get_kind_blade_indices()
    #         or as indices from the blades list property.

    #     Returns:
    #         Geometric algebra torch.Tensor from tensor and blade indices
    #     """
    #     blade_indices = torch.tensor(blade_indices, dtype=torch.int64).to(dtype=torch.int64)
    #     tensor = torch.tensor(tensor, dtype=torch.float32)
    #     # print(f"blade_indices={blade_indices}")
    #     # print(f"tensor={tensor}")
        
    #     _shape = tensor.shape
    #     is_scalar = False
    #     if len(_shape)==1 :
    #         _shape_final = [1]+ [self.num_blades] 
    #         is_scalar = True
    #     else:
    #         _shape_final = list(_shape[:-1]) + [self.num_blades] 
    #     b = torch.zeros(_shape_final)
        

    #     # i = blade_indices.view([-1,1])
    #     # v = tensor.flatten().view([-1,1])
    #     i = blade_indices.nonzero().flatten()
    #     v = tensor.flatten().unsqueeze(1)
    #     b = b.view([-1,self.num_blades])
    #     # b[:,i] = v
    #     try:
    #         b[:,i] = v
    #     except:
    #         print(f"_shape={_shape},_shape_final={_shape_final}")
    #         print(f"i.shape={i.shape},v.shape={v.shape},b.shape={b.shape}")
    #         print(f"i={i},v={v},b={b}")
    #         raise
    #     #     raise "whatever"
    #     b = b.reshape(_shape_final)

    #     # _shape_tmp = list(v.shape) + [self.num_blades] 
    #     # print(f"i,v,_shape_tmp,_shape_final={i},{v},{_shape_tmp},{_shape_final},i.shape={i.shape}")
    #     # b = torch.sparse_coo_tensor(i, v, size=_shape_tmp)
    #     # print(f"b={b}")
    #     # b = torch.sparse_coo_tensor(i, v, size=_shape_tmp).to_dense()
    #     # b = b.reshape(_shape_final)
    #     if is_scalar:
    #         b=b.unsqueeze(0)
    #     return b

    #     # # Put last axis on first axis so scatter_nd becomes easier.
    #     # # Later undo the transposition again.
    #     # # t = tf.concat([[tensor.shape.ndims - 1],
    #     # #                tf.range(0, tensor.shape.ndims - 1)], axis=0)
    #     # # t_inv = tf.concat([tf.range(1, tensor.shape.ndims), [0]], axis=0)

    #     # # tensor = tf.transpose(tensor, t)

    #     # # shape = tf.concat([
    #     # #     torch.tensor([self.num_blades], dtype=torch.int64),
    #     # #     tf.shape(tensor, torch.int64)[1:]
    #     # # ], axis=0)

    #     # # tensor = tf.scatter_nd(
    #     # #     tf.expand_dims(blade_indices, axis=-1),
    #     # #     tensor,
    #     # #     shape
    #     # # )

    #     # # return tf.transpose(tensor, t_inv)
    #     # # t = torch.concat([torch.tensor([len(tensor.shape) - 1]), torch.range(0, len(tensor.shape)- 1)], axis=0)
    #     # # t_inv = torch.concat([torch.range(1, len(tensor.shape)), torch.tensor([0])], axis=0)
    #     # t = [len(tensor.shape) - 1] + list(range(0, len(tensor.shape)- 1))
    #     # t_inv = list(range(1, len(tensor.shape))) +  [0]

    #     # tensor = torch.permute(tensor, t)

    #     # a= torch.tensor([self.num_blades], dtype=torch.int64)
    #     # b = torch.tensor(tensor, dtype=torch.int64)[1:]
    #     # print("a,b:", a,b, tensor)


    #     # shape = torch.concat([
    #     #     torch.tensor([self.num_blades], dtype=torch.int64),
    #     #     torch.tensor(tensor, dtype=torch.int64)[1:]
    #     # ], axis=0)


    #     # # tensor = torch.scatter_nd(
    #     # #     blade_indices.unsqueeze(-1),
    #     # #     tensor,
    #     # #     shape
    #     # # )
    #     # a = torch.zeros(shape)
    #     # a[blade_indices] = tensor
    #     # tensor = a

    #     # return torch.permute(tensor, t_inv) 
            

    def from_tensor(self, tensor: torch.Tensor, blade_indices: torch.Tensor) -> torch.Tensor:
        """Creates a geometric algebra torch.Tensor from a torch.Tensor and blade
        indices. The blade indices have to align with the last axis of the
        tensor.

        Args:
            tensor: torch.Tensor to take as values for the geometric algebra tensor
            blade_indices: Blade indices corresponding to the tensor. Can
            be obtained from blade names eg. using get_kind_blade_indices()
            or as indices from the blades list property.

        Returns:
            Geometric algebra torch.Tensor from tensor and blade indices
        """
        # blade_indices = torch.tensor(blade_indices, dtype=torch.int64).to(dtype=torch.int64)
        # tensor = torch.tensor(tensor, dtype=torch.float32)
        blade_indices = blade_indices.to(dtype=torch.int64)
        tensor = tensor.to(dtype=torch.float32)
        # print(f"blade_indices={blade_indices}")
        # print(f"tensor={tensor}")
        
        _shape = tensor.shape
        is_scalar = False
        if len(_shape)==1 :
            _shape_final = [1]+ [self.num_blades] 
            is_scalar = True
        else:
            _shape_final = list(_shape[:-1]) + [self.num_blades] 
        b = torch.zeros(_shape_final)

        if False:
            print(f"blade_indices.shape={blade_indices.shape}")
            print(f"tensor.shape={tensor.shape}")
            print(f"_shape_final={_shape_final}")
                


        # i = blade_indices.view([-1,1])
        # v = tensor.flatten().view([-1,1])
        # i = blade_indices.nonzero().flatten()
        i = blade_indices.flatten()
        # v = tensor.flatten().unsqueeze(1)
        v = tensor.view([-1,_shape[-1]])
        b = b.view([-1,self.num_blades])
        if False:
            print(f"_shape={_shape},_shape_final={_shape_final}")
            print(f"i.shape={i.shape},v.shape={v.shape},b.shape={b.shape}")
            print(f"i={i},v={v},b={b}")

        # b[:,i] = v
        try:
            b[:,i] = v
        except:
            print(f"_shape={_shape},_shape_final={_shape_final}")
            print(f"i.shape={i.shape},v.shape={v.shape},b.shape={b.shape}")
            print(f"i={i},v={v},b={b}")
            raise
        b = b.reshape(_shape_final)

        if False:
            print(f"b.shape={b.shape}")

        if is_scalar:
            # b=b.unsqueeze(0)
            b=b.squeeze(0)
        return b


        # # i = blade_indices.view([-1,1])
        # # v = tensor.flatten().view([-1,1])
        # i = blade_indices.nonzero().flatten()
        # v = tensor.flatten().unsqueeze(1)
        # b = b.view([-1,self.num_blades])
        # # b[:,i] = v
        # try:
        #     b[:,i] = v
        # except:
        #     print(f"_shape={_shape},_shape_final={_shape_final}")
        #     print(f"i.shape={i.shape},v.shape={v.shape},b.shape={b.shape}")
        #     print(f"i={i},v={v},b={b}")
        #     raise
        # b = b.reshape(_shape_final)

        # if is_scalar:
        #     b=b.unsqueeze(0)
        # return b

       

    def from_tensor_with_kind(self, tensor: torch.Tensor, kind: BladeKind) -> torch.Tensor:
        """Creates a geometric algebra torch.Tensor from a torch.Tensor and a kind.
        The kind's blade indices have to align with the last axis of the
        tensor.

        Args:
            tensor: torch.Tensor to take as values for the geometric algebra tensor
            kind: Kind corresponding to the tensor

        Returns:
            Geometric algebra torch.Tensor from tensor and kind
        """
        # Put last axis on first axis so scatter_nd becomes easier.
        # Later undo the transposition again.
        # tensor = torch.tensor(tensor, dtype=torch.float32)
        tensor = tensor.to(dtype=torch.float32)
        kind_indices = self.get_kind_blade_indices(kind)
        if False:
            print(f"tensor={tensor}")
            print(f"kind_indices={kind_indices}")
        return self.from_tensor(tensor, kind_indices)

    def from_scalar(self, scalar: numbers.Number) -> torch.Tensor:
        """Creates a geometric algebra torch.Tensor with scalar elements.

        Args:
            scalar: Elements to be used as scalars

        Returns:
            Geometric algebra torch.Tensor from scalars
        """
        # return self.from_tensor_with_kind(tf.expand_dims(scalar, axis=-1), BladeKind.SCALAR)
        # print("torch.tensor([scalar]).unsqueeze(-1).shape",torch.tensor([scalar]).unsqueeze(-1).shape)
        return self.from_tensor_with_kind(torch.tensor([scalar]).unsqueeze(-1), BladeKind.SCALAR).squeeze(0)

    def e(self, *blades: List[str]) -> torch.Tensor:
        """Returns a geometric algebra torch.Tensor with the given blades set
        to 1.

        Args:
            blades: list of blade names, can be unnormalized

        Returns:
            torch.Tensor with blades set to 1
        """
        blade_signs, blade_indices = get_blade_indices_from_names(
            blades, self.blades)

        assert type(blade_indices) in [torch.Tensor], "should be a tensor"
        if False: blade_indices = torch.tensor(blade_indices)

        # # Don't allow duplicate indices
        # tf.Assert(
        #     blade_indices.shape[0] == tf.unique(blade_indices)[0].shape[0],
        #     [blades]
        # )

        # x = (
        #     tf.expand_dims(blade_signs, axis=-1) *
        #     tf.gather(self.blade_mvs, blade_indices)
        # )

        # # a, b -> b
        # return tf.reduce_sum(x, axis=-2)

        # print(f"blade_indices={blade_indices}")
        # print(f"torch.unique(blade_indices)={torch.unique(blade_indices)}")
        # print(f"torch.unique(blade_indices)[0]={torch.unique(blade_indices)[0]}")
        # Don't allow duplicate indices
        # assert(
        #     blade_indices.shape[0] == torch.unique(blade_indices).shape[0],
        #     [blades]
        # )
        assert blade_indices.shape[0] == torch.unique(blade_indices).shape[0], "indexes not unique"

        x = blade_signs.unsqueeze(-1) *  self.blade_mvs[blade_indices]

        # a, b -> b
        return x.sum(dim=-2)        

    def __getattr__(self, name: str) -> torch.Tensor:
        """Returns basis blade tensors if name was a basis."""
        if name.startswith("e") and (name[1:] == "" or int(name[1:]) >= 0):
            return self.e(name[1:])
        raise AttributeError

    def dual(self, tensor: torch.Tensor) -> torch.Tensor:
        """Returns the dual of the geometric algebra tensor.

        Args:
            tensor: Geometric algebra tensor to return dual for

        Returns:
            Dual of the geometric algebra tensor
        """
        tensor = torch.tensor(tensor, dtype=torch.float32)
        # return self.dual_blade_signs * tf.gather(tensor, self.dual_blade_indices, axis=-1)
        return self.dual_blade_signs * tensor[...,self.dual_blade_indices]

    def grade_automorphism(self, tensor: torch.Tensor) -> torch.Tensor:
        """Returns the geometric algebra tensor with odd grades negated.
        See https://en.wikipedia.org/wiki/Paravector#Grade_automorphism.

        Args:
            tensor: Geometric algebra tensor to return grade automorphism for

        Returns:
            Geometric algebra tensor with odd grades negated
        """
        tensor = tensor.to(dtype=torch.float32)
        return mv_grade_automorphism(tensor, self.blade_degrees)

    def reversion(self, tensor: torch.Tensor) -> torch.Tensor:
        """Returns the grade-reversed geometric algebra tensor.
        See https://en.wikipedia.org/wiki/Paravector#Reversion_conjugation.

        Args:
            tensor: Geometric algebra tensor to return grade-reversion for

        Returns:
            Grade-reversed geometric algebra tensor
        """
        tensor = tensor.to(dtype=torch.float32)

        return mv_reversion(tensor, self.blade_degrees)

    def conjugation(self, tensor: torch.Tensor) -> torch.Tensor:
        """Combines reversion and grade automorphism.
        See https://en.wikipedia.org/wiki/Paravector#Clifford_conjugation.

        Args:
            tensor: Geometric algebra tensor to return conjugate for

        Returns:
            Geometric algebra tensor after `reversion()` and `grade_automorphism()`
        """
        tensor = tensor.to(dtype=torch.float32)
        return self.grade_automorphism(self.reversion(tensor))

    def simple_inverse(self, a: torch.Tensor) -> torch.Tensor:
        """Returns the inverted geometric algebra tensor
        `X^-1` such that `X * X^-1 = 1`. Only works for elements that
        square to scalars. Faster than the general inverse.

        Args:
            a: Geometric algebra tensor to return inverse for

        Returns:
            inverted geometric algebra tensor
        """
        a = a.to(dtype=torch.float32)


        rev_a = self.reversion(a)
        divisor = self.geom_prod(a, rev_a)
        # print(f"divisor={divisor}")
        # print(f"self.is_pure_kind(divisor, BladeKind.SCALAR)={self.is_pure_kind(divisor, BladeKind.SCALAR)}")
        if not self.is_pure_kind(divisor, BladeKind.SCALAR):
            raise Exception(
                "Can't invert multi-vector (inversion divisor V ~V not scalar: %s)." % divisor)

        # Divide by scalar part
        return rev_a / divisor[..., :1]

    def reg_prod(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        """Returns the regressive product of two geometric
        algebra tensors.

        Args:
            a: Geometric algebra tensor on the left hand side of
            the regressive product
            b: Geometric algebra tensor on the right hand side of
            the regressive product

        Returns:
            regressive product of a and b
        """
        a = torch.tensor(a, dtype=torch.float32)
        b = torch.tensor(b, dtype=torch.float32)

        return self.dual(self.ext_prod(self.dual(a), self.dual(b)))

    def ext_prod(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        """Returns the exterior product of two geometric
        algebra tensors.

        Args:
            a: Geometric algebra tensor on the left hand side of
            the exterior product
            b: Geometric algebra tensor on the right hand side of
            the exterior product

        Returns:
            exterior product of a and b
        """
        a = a.to(dtype=torch.float32)
        b = b.to(dtype=torch.float32)

        return mv_multiply(a, b, self._cayley_outer)

    def geom_prod(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        """Returns the geometric product of two geometric
        algebra tensors.

        Args:
            a: Geometric algebra tensor on the left hand side of
            the geometric product
            b: Geometric algebra tensor on the right hand side of
            the geometric product

        Returns:
            geometric product of a and b
        """
        # a = torch.tensor(a, dtype=torch.float32)
        # b = torch.tensor(b, dtype=torch.float32)

        # a = torch.tensor(a)
        # b = torch.tensor(b)

        a = a.to(dtype=torch.float32)
        b = b.to(dtype=torch.float32)
        return mv_multiply(a, b, self._cayley)

    
    def element_wise_prod(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        """Returns the element-wise product of two geometric
        algebra tensors.

        Args:
            a: Geometric algebra tensor on the left hand side of
            the geometric product
            b: Geometric algebra tensor on the right hand side of
            the geometric product

        Returns:
            geometric product of a and b
        """
        # a = torch.tensor(a, dtype=torch.float32)
        # b = torch.tensor(b, dtype=torch.float32)

        # a = torch.tensor(a)
        # b = torch.tensor(b)

        a = a.to(dtype=torch.float32)
        b = b.to(dtype=torch.float32)
        return mv_multiply_element_wise(a, b, self._cayley)


    def inner_prod(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        """Returns the inner product of two geometric
        algebra tensors.

        Args:
            a: Geometric algebra tensor on the left hand side of
            the inner product
            b: Geometric algebra tensor on the right hand side of
            the inner product

        Returns:
            inner product of a and b
        """
        a = a.to(dtype=torch.float32)
        b = b.to(dtype=torch.float32)

        return mv_multiply(a, b, self._cayley_inner)

    def geom_conv1d(self, a: torch.Tensor, k: torch.Tensor,
                    stride: int, padding: str,
                    dilations: Union[int, None] = None) -> torch.Tensor:
        """Returns the 1D convolution of a sequence with a geometric algebra
        tensor kernel. The convolution is performed using the geometric
        product.

        Args:
            a: Input geometric algebra tensor of shape
                [..., Length, ChannelsIn, Blades]
            k: Geometric algebra tensor for the convolution kernel of shape
                [KernelSize, ChannelsIn, ChannelsOut, Blades]
            stride: Stride to use for the convolution
            padding: "SAME" (zero-pad input length so output
                length == input length / stride) or "VALID" (no padding)
        Returns:
            Geometric algbra tensor of shape
            [..., OutputLength, ChannelsOut, Blades]
            representing `a` convolved with `k`
        """
        a = a.to(dtype=torch.float32)
        k = k.to(dtype=torch.float32)

        # return mv_conv1d(a, k, self._cayley, stride=stride, padding=padding)
        return f_mv_conv1d(a, k, self._cayley, stride=stride, padding=padding)

    def mv_repr(self, a: torch.Tensor) -> str:
        """Returns a string representation for the given
        geometric algebra tensor.

        Args:
            a: Geometric algebra tensor to return the representation for

        Returns:
            string representation for `a`
        """
        a = a.to(dtype=torch.float32)


        if len(a.shape) == 1:
            return "MultiVector[%s]" % " + ".join(
                "%.2f*%s" % (value, get_blade_repr(blade_name))
                for value, blade_name
                in zip(a, self.blades)
                if value != 0
            )
        else:
            return f"MultiVector[batch_shape={a.shape[:-1]}]"

    def approx_exp(self, a: torch.Tensor, order: int = 50) -> torch.Tensor:
        """Returns an approximation of the exponential using a centered taylor series.

        Args:
            a: Geometric algebra tensor to return exponential for
            order: order of the approximation

        Returns:
            Approximation of `exp(a)`
        """
        a = a.to(dtype=torch.float32)

        v = self.from_scalar(1.0)
        result = self.from_scalar(1.0)
        for i in range(1, order + 1):
            v = self.geom_prod(a, v)
            # i_factorial = tf.exp(tf.math.lgamma(i + 1.0))
            i_factorial = torch.exp(torch.lgamma(torch.tensor([i + 1.0])))
            result += v / i_factorial
        return result

    def exp(self, a: torch.Tensor, square_scalar_tolerance: Union[float, None] = 1e-4) -> torch.Tensor:
        """Returns the exponential of the passed geometric algebra tensor.
        Only works for multivectors that square to scalars.

        Args:
            a: Geometric algebra tensor to return exponential for
            square_scalar_tolerance: Tolerance to use for the square scalar check
                or None if the check should be skipped

        Returns:
            `exp(a)`
        """
        # See https://www.euclideanspace.com/maths/algebra/clifford/algebra/functions/exponent/index.htm
        # for an explanation of how to exponentiate multivectors.

        self_sq = self.geom_prod(a, a)

        if square_scalar_tolerance is not None:
            # tf.Assert(tf.reduce_all(
            #     tf.abs(self_sq[..., 1:]) < square_scalar_tolerance
            # ), [self_sq])
            
            # assert torch.equal(torch.all(self_sq[..., 1:].abs() < square_scalar_tolerance),[self_sq]), "not sure what"
            assert torch.all(self_sq[..., 1:].abs() < square_scalar_tolerance), "square_scalar_tolerance not met"

        scalar_self_sq = self_sq[..., :1]

        # "Complex" square root (argument can be negative)
        s_sqrt = torch.sign(scalar_self_sq) * torch.sqrt(torch.abs(scalar_self_sq))

        # Square to +1: cosh(sqrt(||a||)) + a / sqrt(||a||) sinh(sqrt(||a||))
        # Square to -1: cos(sqrt(||a||)) + a / sqrt(||a||) sin(sqrt(||a||))
        # TODO: Does this work for values other than 1 too? eg. square to +0.5?
        # TODO: Find a solution that doesnt require calculating all possibilities
        #       first.
        non_zero_result = torch.where(
            scalar_self_sq < 0,
            (self.from_tensor(torch.cos(s_sqrt), torch.tensor([0])) +  a / s_sqrt * torch.sin(s_sqrt)),
            (self.from_tensor(torch.cosh(s_sqrt), torch.tensor([0])) +  a / s_sqrt * torch.sinh(s_sqrt))
        )

        return torch.where(scalar_self_sq == 0, self.from_scalar(1.0) + a, non_zero_result)

    def approx_log(self, a: torch.Tensor, order: int = 50) -> torch.Tensor:
        """Returns an approximation of the natural logarithm using a centered
        taylor series. Only converges for multivectors where `||mv - 1|| < 1`.

        Args:
            a: Geometric algebra tensor to return logarithm for
            order: order of the approximation

        Returns:
            Approximation of `log(a)`
        """
        a = a.to(dtype=torch.float32)

        result = self.from_scalar(0.0)

        a_minus_one = a - self.from_scalar(1.0)
        v = None

        for i in range(1, order + 1):
            v = a_minus_one if v is None else v * a_minus_one
            result += (((-1.0) ** i) / i) * v

        return -result

    def int_pow(self, a: torch.Tensor, n: int) -> torch.Tensor:
        """Returns the geometric algebra tensor to the power of an integer
        using repeated multiplication.

        Args:
            a: Geometric algebra tensor to raise
            n: integer power to raise the multivector to

        Returns:
            `a` to the power of `n`
        """
        a = a.to(dtype=torch.float32)


        if not isinstance(n, int):
            raise Exception("n must be an integer.")
        if n < 0:
            raise Exception("Can't raise to negative powers.")

        if n == 0:
            # TODO: more efficient (ones only in scalar)
            return torch.ones_like(a) * self.e("")

        result = a
        for i in range(n - 1):
            result = self.geom_prod(result, a)
        return result

    def keep_blades(self, a: torch.Tensor, blade_indices: List[int]) -> torch.Tensor:
        """Takes a geometric algebra tensor and returns it with only the given
        blade_indices as non-zeros.

        Args:
            a: Geometric algebra tensor to copy
            blade_indices: Indices for blades to keep

        Returns:
            `a` with only `blade_indices` components as non-zeros
        """
        a = a.to(dtype=torch.float32)
        blade_indices = blade_indices.to(dtype=torch.int64)

        # blade_values = tf.gather(a, blade_indices, axis=-1)
        blade_values = a[...,blade_indices]
        if True: 
            b = self.from_tensor(blade_values, blade_indices)
        else:
            blade_mask = torch.zeros(self.num_blades)
            blade_mask[blade_indices] = 1
            b = self.from_tensor(blade_values, blade_mask)
        # print(f"blade_values, blade_indices, b={blade_values}, {blade_indices}, {b}")
        # print(f"blade_mask={blade_mask}")
        return b

        # return self.from_tensor(blade_values, blade_indices)

    def keep_blades_with_name(self, a: torch.Tensor, blade_names: Union[List[str], str]) -> torch.Tensor:
        """Takes a geometric algebra tensor and returns it with only the given
        blades as non-zeros.

        Args:
            a: Geometric algebra tensor to copy
            blade_names: Blades to keep

        Returns:
            `a` with only `blade_names` components as non-zeros
        """
        if isinstance(blade_names, str):
            blade_names = [blade_names]

        _, blade_indices = get_blade_indices_from_names(blade_names, self.blades)

        if False:
            print(f"self.blades={self.blades}")
            print(f"blade_names={blade_names}")
            print(f"blade_indices={blade_indices}")

        return self.keep_blades(a, blade_indices)

    def select_blades(self, a: torch.Tensor, blade_indices: List[int]) -> torch.Tensor:
        """Takes a geometric algebra tensor and returns a `torch.Tensor` with the
        blades in blade_indices on the last axis.


        Args:
            a: Geometric algebra tensor to copy
            blade_indices: Indices for blades to select

        Returns:
            `torch.Tensor` based on `a` with `blade_indices` on last axis.
        """
        a = a.to(dtype=torch.float32)        
        # blade_indices = torch.tensor(blade_indices, dtype=torch.int64).to(dtype=torch.int64)
        blade_indices = blade_indices.to(dtype=torch.int64)

        # result = tf.gather(a, blade_indices, axis=-1)
        try:
            if len(a.shape)==1 or a.shape[-1]==a.size().numel():
                result = a.squeeze()[blade_indices]
            else:
                result = a[...,blade_indices]
        except:
            print(f"a={a},blade_indices={blade_indices}")
            print(f"a.shape={a.shape},blade_indices.shape={blade_indices.shape},a.size().numel()={a.size().numel()}")
            raise
        
        return result

    def select_blades_with_name(self, a: torch.Tensor, blade_names: Union[List[str], str]) -> torch.Tensor:
        """Takes a geometric algebra tensor and returns a `torch.Tensor` with the
        blades in blade_names on the last axis.


        Args:
            a: Geometric algebra tensor to copy
            blade_names: Blades to keep

        Returns:
            `torch.Tensor` based on `a` with `blade_names` on last axis.
        """
        a = a.to(dtype=torch.float32)

        is_single_blade = isinstance(blade_names, str)
        if is_single_blade:
            blade_names = [blade_names]

        blade_signs, blade_indices = get_blade_indices_from_names(
            blade_names, self.blades)

        result = blade_signs * self.select_blades(a, blade_indices)
        # if True:
        #     print(f"")

        if is_single_blade:
            return result[..., 0]

        return result

    def inverse(self, a: torch.Tensor) -> torch.Tensor:
        """Returns the inverted geometric algebra tensor
        `X^-1` such that `X * X^-1 = 1`.

        Using Shirokov's inverse algorithm that works in arbitrary dimensions,
        see https://arxiv.org/abs/2005.04015 Theorem 4.

        Args:
            a: Geometric algebra tensor to return inverse for

        Returns:
            inverted geometric algebra tensor
        """
        # a = torch.tensor(a, dtype=torch.float32)
        a = a.to(dtype=torch.float32)
        if False:
            print(f"a={a}")

        n = 2 ** ((len(self.metric) + 1) // 2)

        # u = a.clone()
        u = a
        for k in range(1, n):
            # c = n / k * self.keep_blades_with_name(u, "")
            d = self.keep_blades_with_name(u, "")
            c = n / k * d
            u_minus_c = u - c
            if False:
                print(f"a,d,c,u_minus_c, u = {a},{d},{c},{u_minus_c}, {u}")
            u = self.geom_prod(a, u_minus_c)
            if False:
                print(f"u={u}")
        
        if False:
            print(f"n={n}")
            print(f"a={a}")
            print(f"u={u}")
        if not torch.all(self.is_pure_kind(u, BladeKind.SCALAR)):
            raise Exception(
                "Can't invert multi-vector (det U not scalar: %s)." % u)

        # adj / det
        return u_minus_c / u[..., :1]

    def __call__(self, a: torch.Tensor) -> MultiVector:
        """Creates a `MultiVector` from a geometric algebra tensor.
        Mainly used as a wrapper for the algebra's functions for convenience.

        Args:
            a: Geometric algebra tensor to return `MultiVector` for

        Returns:
            `MultiVector` for `a`
        """
        a = a.to(dtype=torch.float32)
        return MultiVector(a, self)
        # return MultiVector(torch.tensor(a), self)



"""Operations on geometric algebra tensors used internally."""
from typing import Union
import torch


def mv_multiply(a_blade_values: torch.Tensor, b_blade_values: torch.Tensor, cayley: torch.Tensor) -> torch.Tensor:
    # x = torch.einsum("i,j,ijk->k", a_blade_values, b_blade_values, cayley)
    

    # cehck later
    # # # ...i, ijk -> ...jk
    # # x = torch.tensordot(a_blade_values, cayley, dims=[-1, 0])
    # x = torch.tensordot(a_blade_values, cayley, dims=([-1, 0],[-1,0]))
    # # # ...1j, ...jk -> ...1k
    # # x = tf.expand_dims(b_blade_values, axis=b_blade_values.shape.ndims - 1) @ x
    # x = b_blade_values.unsqueeze(len(b_blade_values.shape) - 1) @ x
    # # # ...1k -> ...k
    # # x = torch.squeeze(x, axis=-2)
    # x = torch.squeeze(x, axis=-2)
    
    # cehck later
    # # ...i, ijk -> ...jk
    # x = torch.tensordot(a_blade_values, cayley, dims=[-1, 0])
    # x = torch.tensordot(a_blade_values, cayley, dims=([-1, 0],[-1,0]))
    x = torch.tensordot(a_blade_values, cayley, dims=([-1],[0]))

    # # ...1j, ...jk -> ...1k
    # x = tf.expand_dims(b_blade_values, axis=b_blade_values.shape.ndims - 1) @ x
    # x = b_blade_values.unsqueeze(len(b_blade_values.shape) - 1) @ x
    x = b_blade_values.unsqueeze(-2) @ x        
    # # ...1k -> ...k
    # x = torch.squeeze(x, axis=-2)
    x = torch.squeeze(x, axis=-2) 
       
    # # # ...1j, ...jk -> ...1k
    # x = b_blade_values @ x        
    
    # print(f"same opeartions? x.shape={x.shape},x1.shape={x1.shape}")
    
    # # einsum
    # x1 = torch.einsum("...i,...j,ijk->...k", a_blade_values, b_blade_values, cayley)    
    # assert(torch.all(torch.isclose(x1,x))), f"should be the same operation x[0]={x[0]}, x1[0]={x1[0]}"
    

    return x


def mv_multiply_element_wise(a_blade_values: torch.Tensor, b_blade_values: torch.Tensor, cayley: torch.Tensor) -> torch.Tensor:
    x = a_blade_values * b_blade_values
    return x

import math
import torch.nn.functional as F
# https://discuss.pytorch.org/t/tf-extract-image-patches-in-pytorch/43837/10
def extract_image_patches(x, kernel, stride=1, dilation=1):
    # Do TF 'SAME' Padding
    b,c,h,w = x.shape
    h2 = math.ceil(h / stride)
    w2 = math.ceil(w / stride)
    pad_row = (h2 - 1) * stride + (kernel - 1) * dilation + 1 - h
    pad_col = (w2 - 1) * stride + (kernel - 1) * dilation + 1 - w
    # x = F.pad(x, (pad_row//2, pad_row - pad_row//2, pad_col//2, pad_col - pad_col//2))
    x = F.pad(x, (pad_col//2, pad_col - pad_col//2, pad_row//2, pad_row - pad_row//2, ))
    
    # Extract patches
    patches = x.unfold(2, kernel, stride).unfold(3, kernel, stride)
    patches = patches.permute(0,4,5,1,2,3).contiguous()
    
    return patches.view(b,-1,patches.shape[-2], patches.shape[-1])

# https://discuss.pytorch.org/t/conv1d-implementation-using-torch-nn-functional-unfold/109643/3
import torch.nn.functional as F
def f_mv_conv1d(input, weight, cayley: torch.Tensor, bias=None, stride=1, padding=0, dilation=1, groups=1):
    """
    input : input tensor of shape    : (minibatch,in_channels, iW)
    input : input tensor of shape    : (minibatch,in_channels, width, num_blades)
    weight : filters of shape : (out_channels, in_channels/groups, kW)
    weight : filters of shape : (out_channels, in_channels/groups, kernel_size, num_blades)
    bias :  optional bias of shape (out_channels). Default: None
    bias :  optional bias of shape (out_channels, num_blades). Default: None
    stride :  the stride of the convolving kernel. Can be a single number or a one-element tuple (sW,). Default: 1
    padding : implicit paddings on both sides of the input. Can be a string {‘valid’, ‘same’}, single number or a one-element tuple (padW,). Default: 0 padding='valid' is the same as no padding. padding='same' pads the input so the output has the same shape as the input. However, this mode doesn’t support any stride values other than 1. 
    dilation : the spacing between kernel elements. Can be a single number or a one-element tuple (dW,). Default: 1
    groups : split input into groups, in_channels should be divisible by the number of groups. Default: 1    
    """
    # kernel_size = weight.shape

    assert len(input.shape)==4, "input size == 4 (minibatch,in_channels, width, num_blades)"
    assert len(weight.shape)==4, "weights size == 4 (out_channels, in_channels/groups, kernel_size, num_blades)"

    # A: [..., S, CI, BI]
    # K: [K, CI, CO, BK]
    # C: [BI, BK, BO]    
    input = input.permute(0,2,1,3)
    weight = weight.permute(2,1,0,3)

    batch,in_channels,width,num_blades = input.shape
    out_channels, in_channels, kernel,num_blades1 = weight.shape
    assert (num_blades==num_blades1), "same geometry please"
    kernel_size = (kernel,num_blades)

    input_unfold = F.unfold(input.view(batch * groups, in_channels // groups, width, num_blades), kernel_size, dilation, 0, stride)
    # N,Cxprod_kernel,L
    # input_unfold = input_unfold.view(batch, groups, input_unfold.size(1), input_unfold.size(2))
    input_unfold = input_unfold.view(batch, in_channels // groups, kernel, num_blades, input_unfold.size(2))
    # ci,ks,bi,L * co,ci,ks,bj * bi,bj,bk -> co,ks,L,bk  
    # a,b,c,d * e,a,b,f * c,f,g -> e,b,d,g
    # ...abcd, eabf, cfg -> ...ebdg  
    x = torch.einsum("...abcd, eabf, cfg -> ...ebdg", input_unfold, weight, cayley)
    x = x.view(batch,out_channels,-1,num_blades) + (bias.view(1,out_channels,1,num_blades) if bias else 0) 
    x = x.permute(0,2,1,3)
    return x

    # input = input.unqueeze(3) #now size is 4
    # input_unfold = F.unfold(input, kernel_size, dilation, padding, stride)
    # out_unfold = input_unfold.transpose(1, 2).matmul(weight.view(weight.size(0), -1).t()).transpose(1, 2)
    # # input, output_size, kernel_size, dilation=1, padding=0, stride=1
    # F.fold(out_unfold, output_size, (1, 1),dilation, padding, stride)
    

import numpy as np
def mv_conv1d(a_blade_values: torch.Tensor, k_blade_values: torch.Tensor, cayley: torch.Tensor,
              stride: int, padding: str, dilations: Union[int, None] = None) -> torch.Tensor:
    # Winograd convolution

    # A: [..., S, CI, BI]
    # K: [K, CI, CO, BK]
    # C: [BI, BK, BO]

    kernel_size = k_blade_values.shape[0]

    a_batch_shape = a_blade_values.shape[:-3]

    # Reshape a_blade_values to a 2d image (since that's what the tf op expects)
    # [*, S, 1, CI*BI]
    # a_image_shape = torch.concat([
    #     torch.tensor(a_batch_shape),
    #     torch.tensor(a_blade_values.shape[-3:-2]),
    #     torch.tensor([1, torch.prod(torch.tensor(a_blade_values.shape[-2:]))])
    # ], axis=0)
    a_image_shape = list(a_batch_shape) + list(a_blade_values.shape[-3:-2]) + [1, np.prod(a_blade_values.shape[-2:]) ]
    print(f"a_image_shape={a_image_shape}")
    a_image = torch.reshape(a_blade_values, tuple([int(_) for _ in a_image_shape]))

    sizes = [1, kernel_size, 1, 1]
    strides = [1, stride, 1, 1]

    # [*, P, 1, K*CI*BI] where eg. number of patches P = S * K for
    # stride=1 and "SAME", (S-K+1) * K for "VALID", ...
    # a_slices = tf.image.extract_patches(
    #     a_image,
    #     sizes=sizes, strides=strides,
    #     rates=[1, 1, 1, 1], padding=padding
    # )
    # extract_image_patches(x, kernel, stride=1, dilation=1):
    a_slices = extract_image_patches(
        a_image,
        sizes, stride=strides
        # rates=[1, 1, 1, 1], 
        # padding=padding
    )

    # https://pytorch.org/docs/stable/generated/torch.nn.Unfold.html
    # inp_unf = F.unfold(a_image, kernel_size=sizes, dilation=1, padding=padding, stride=strides)
    # out_unf = inp_unf.transpose(1, 2).matmul(w.view(w.size(0), -1).t()).transpose(1, 2)
    # out = F.fold(out_unf, (7, 8), (1, 1))

    # [..., P, K, CI, BI]
    out_shape = torch.concat([
        a_batch_shape,
        a_slices.shape()[-3:-2],
        k_blade_values.shape()[:1],
        a_blade_values.shape()[-2:]
    ], axis=0)

    a_slices = torch.reshape(a_slices, out_shape)

    # TODO: Optimize this to not use einsum (since it's slow with ellipses)
    # a_...p,k,ci,bi; k_k,ci,co,bk; c_bi,bk,bo -> y_...p,co,bo
    #   ...a b c  d ,   e c  f  g ,   d  g  h  ->   ...a f  h
    x = torch.einsum("...abcd,bcfg,dgh->...afh", a_slices, k_blade_values, cayley)
    return x


def mv_reversion(a_blade_values, algebra_blade_degrees):
    algebra_blade_degrees = algebra_blade_degrees.to(torch.float32)
    # for each blade, 0 if even number of swaps required, else 1
    odd_swaps =  (torch.floor(algebra_blade_degrees * (algebra_blade_degrees - 0.5)) % 2).to(dtype=torch.float32)
    # [0, 1] -> [-1, 1]
    reversion_signs = 1.0 - 2.0 * odd_swaps
    return reversion_signs * a_blade_values


def mv_grade_automorphism(a_blade_values, algebra_blade_degrees):
    algebra_blade_degrees = algebra_blade_degrees.to(dtype=torch.float32)
    signs = 1.0 - 2.0 * (algebra_blade_degrees % 2.0)
    return signs * a_blade_values



"""Provides Geometric Algebra Keras layers."""
from typing import List, Union


import numpy as np
import torch
import torch.nn as nn
import torch.nn.init as init

from .blades import BladeKind

from .torch_ga import GeometricAlgebra

class GeometricAlgebraLayer(nn.Module):
    def __init__(self, algebra: GeometricAlgebra, **kwargs):
        self.algebra = algebra
        self.built = False
        super().__init__(**kwargs)
    
    def build(self):
        assert(False), "why me?"

    @classmethod
    def from_config(cls, config):
        # Create algebra if necessary (should only occur once, assumes that
        # config is actually mutable).
        if "algebra" not in config:
            assert "metric" in config
            config["algebra"] = GeometricAlgebra(config["metric"])
            del config["metric"]
        return cls(**config)

    def get_config(self):
        # Store metric of the algebra. In from_config() we will recreate the
        # algebra from the metric.
        config = super().get_config()
        config.update({
            "metric": self.algebra.metric.numpy()
        })
        return config


# @register_keras_serializable(package="TFGA")
class TensorToGeometric(GeometricAlgebraLayer):
    """Layer for converting tensors with given blade indices to
    geometric algebra tensors.

    Args:
        algebra: GeometricAlgebra instance to use
        blade_indices: blade indices to interpret the last axis of the
        input tensor as
    """

    def __init__(self, algebra: GeometricAlgebra, blade_indices: List[int],
                 **kwargs):
        super().__init__(algebra=algebra, **kwargs)

        self.blade_indices = torch.tensor(blade_indices, dtype=torch.int64)
        # self.blade_indices = blade_indices.to(dtype=torch.int64)     
        self.built = False

    def compute_output_shape(self, input_shape):
        return [*input_shape[:-1], self.algebra.num_blades]

    def forward(self, inputs):
        if not self.build: self.build(inputs.shape)
        return self.algebra.from_tensor(inputs, blade_indices=self.blade_indices)
    def build(self,input_shape):  self.built = True
    def get_config(self):
        config = super().get_config()
        config.update({
            "blade_indices": self.blade_indices.numpy()
        })
        return config


# @register_keras_serializable(package="TFGA")
class TensorWithKindToGeometric(GeometricAlgebraLayer):
    """Layer for converting tensors with given blade kind to
    geometric algebra tensors.

    Args:
        algebra: GeometricAlgebra instance to use
        kind: blade kind indices to interpret the last axis of the
        input tensor as
    """

    def __init__(self, algebra: GeometricAlgebra, kind: BladeKind,
                 **kwargs):
        super().__init__(algebra=algebra, **kwargs)
        self.kind = kind
        self.built = False

    def compute_output_shape(self, input_shape):
        return [*input_shape[:-1], self.algebra.get_kind_blade_indices(self.kind).shape[0]]

    def build(self,input_shape): self.built = True
    def forward(self, inputs):
        if not self.build: self.build(inputs.shape)

        return self.algebra.from_tensor_with_kind(inputs, kind=self.kind)

    def get_config(self):
        config = super().get_config()
        config.update({
            "kind": self.kind
        })
        return config


# @register_keras_serializable(package="TFGA")
class GeometricToTensor(GeometricAlgebraLayer):
    """Layer for extracting given blades from geometric algebra tensors.

    Args:
        algebra: GeometricAlgebra instance to use
        blade_indices: blade indices to extract
    """

    def __init__(self, algebra: GeometricAlgebra, blade_indices: List[int],
                 **kwargs):
        super().__init__(algebra=algebra, **kwargs)
        self.blade_indices = torch.tensor(blade_indices).to(dtype=torch.int64)
        # self.blade_indices = blade_indices.to(dtype=torch.int64) 
        self.built = False

    def compute_output_shape(self, input_shape):
        return [*input_shape[:-1], self.blade_indices.shape[0]]
    def build(self,input_shape): self.built = True

    def forward(self, inputs):
        if not self.build: self.build(inputs.shape)
        # return torch.select(inputs, self.blade_indices, axis=-1)
        x = inputs[...,self.blade_indices]
        return x

    def get_config(self):
        config = super().get_config()
        config.update({
            "blade_indices": self.blade_indices.numpy()
        })
        return config


# @register_keras_serializable(package="TFGA")
class GeometricToTensorWithKind(GeometricToTensor):
    """Layer for extracting blades of a kind from geometric algebra tensors.

    Args:
        algebra: GeometricAlgebra instance to use
        kind: blade indices of kind to extract
    """

    def __init__(self, algebra: GeometricAlgebra, kind: BladeKind,
                 **kwargs):
        blade_indices = algebra.get_kind_blade_indices(kind)
        super().__init__(algebra=algebra, blade_indices=blade_indices,
                         **kwargs)



activations = {
    'None': None,
    'relu': nn.ReLU(),
    'sigmoid': nn.Sigmoid(),
    'tanh': nn.Tanh(),
}

initializers = {
    'uniform_': init.uniform_ ,
    'normal_': init.normal_ ,
    'constant_': init.constant_ ,
    'xavier_uniform_': init.xavier_uniform_ ,
    'xavier_normal_': init.xavier_normal_ ,
    'kaiming_uniform_': init.kaiming_uniform_ ,
    'kaiming_normal_': init.kaiming_normal_ ,            
}

# regularizers ={
#     '':,
# }


# from torch.nn.parameter import Parameter

# @register_keras_serializable(package="TFGA")
class GeometricProductDense(GeometricAlgebraLayer):
    """Analagous to Keras' Dense layer but using multivector-valued matrices
    instead of scalar ones and geometric multiplication instead of standard
    multiplication.

    Args:
        algebra: GeometricAlgebra instance to use for the parameters
        blade_indices_kernel: Blade indices to use for the kernel parameter
        blade_indices_bias: Blade indices to use for the bias parameter (if used)
    """

    def __init__(
        self,
        algebra: GeometricAlgebra,
        units: int,
        blade_indices_kernel: List[int],
        blade_indices_bias: Union[None, List[int]] = None,
        activation='None',
        use_bias=True,
        **kwargs
    ):
        super().__init__(algebra=algebra, **kwargs)

        self.units = units
        self.blade_indices_kernel = torch.tensor(blade_indices_kernel, dtype=torch.int64)
        if use_bias: self.blade_indices_bias = torch.tensor(blade_indices_bias, dtype=torch.int64)
        # self.blade_indices_kernel = blade_indices_kernel.to(dtype=torch.int64)
        # if use_bias: self.blade_indices_bias = blade_indices_bias.to(dtype=torch.int64)            

        self.activation = activation
        self.use_bias = use_bias
        self.activation_fn = activations.get(activation)
        self.built = False

    def build(self, input_shape: list):
        if False: print(f"input_shape={input_shape}")
        self.num_input_units = input_shape[-2]
        shape_kernel = [
            self.units,
            self.num_input_units,
            int(self.blade_indices_kernel.shape[0])
        ]
        if False: print(f"shape_kernel={shape_kernel}")
        self.kernel = nn.Parameter(1./np.prod(shape_kernel)*torch.randn(size=shape_kernel)).to(dtype=torch.float)
        if self.use_bias:
            shape_bias = [self.units, self.blade_indices_bias.shape[0]]
            self.bias = nn.Parameter(1./np.prod(shape_bias)*torch.randn(size=shape_bias)).to(dtype=torch.float)
        else:
            self.bias = None
        self.built = True

    def compute_output_shape(self, input_shape):
        return [*input_shape[:-2], self.units, self.algebra.num_blades]

    def forward(self, inputs):
        if not self.built: self.build(inputs.shape)
        w_geom = self.algebra.from_tensor(self.kernel, self.blade_indices_kernel)

        # Perform a matrix-multiply, but using geometric product instead of
        # standard multiplication. To do this we do the geometric product
        # elementwise and then sum over the common axis.
        # [..., 1, I, X] * [..., O, I, X] -> [..., O, I, X] -> [..., O, X]
        # inputs_expanded = tf.expand_dims(inputs, axis=inputs.shape.ndims - 2)
        # result = tf.reduce_sum(self.algebra.geom_prod(
        #     inputs_expanded, w_geom), axis=-2)

        inputs_expanded = inputs.unsqueeze(len(inputs.shape) - 2)
        result = self.algebra.geom_prod(inputs_expanded, w_geom).sum(dim=-2)
        if self.bias is not None:
            b_geom = self.algebra.from_tensor(self.bias, self.blade_indices_bias)
            result += b_geom
        if self.activation_fn:
            result = self.activation_fn(result)
        return result

    def get_config(self):
        config = super().get_config()
        config.update({
            "blade_indices_kernel":
                self.blade_indices_kernel.cpu().detach().numpy(),
            "blade_indices_bias":
                self.blade_indices_bias.cpu().detach().numpy(),
            "units":
                self.units,
            # "activation":
            #     activations.serialize(self.activation),
            "use_bias":
                self.use_bias,
        })
        return config


# @register_keras_serializable(package="TFGA")
class GeometricSandwichProductDense(GeometricProductDense):
    """Analagous to Keras' Dense layer but using multivector-valued matrices
    instead of scalar ones and geometric sandwich multiplication instead of
    standard multiplication.

    Args:
        algebra: GeometricAlgebra instance to use for the parameters
        blade_indices_kernel: Blade indices to use for the kernel parameter
        blade_indices_bias: Blade indices to use for the bias parameter (if used)
    """

    def __init__(
        self, algebra, units, blade_indices_kernel, blade_indices_bias=None,
        activation=None, use_bias=True, 
        # kernel_initializer="glorot_uniform",
        # bias_initializer="zeros", kernel_regularizer=None,
        # bias_regularizer=None, activity_regularizer=None,
        # kernel_constraint=None, bias_constraint=None, 
        **kwargs
    ):
        super().__init__(
            algebra, units,
            blade_indices_kernel,
            blade_indices_bias=blade_indices_bias,
            activation=activation,
            use_bias=use_bias,
            # kernel_initializer=kernel_initializer,
            # bias_initializer=bias_initializer,
            # kernel_regularizer=kernel_regularizer,
            # bias_regularizer=bias_regularizer,
            # activity_regularizer=activity_regularizer,
            # kernel_constraint=kernel_constraint,
            # bias_constraint=bias_constraint, 
            **kwargs
        )
        self.built = False

    def forward(self, inputs):
        if not self.built: self.build(inputs.shape)
        w_geom = self.algebra.from_tensor(self.kernel, self.blade_indices_kernel)

        # Same as GeometricProductDense but using R*x*~R instead of just R*x
        # inputs_expanded = tf.expand_dims(inputs, axis=inputs.shape.ndims - 2)
        # result = tf.reduce_sum(
        #     self.algebra.geom_prod(
        #         w_geom,
        #         self.algebra.geom_prod(
        #             inputs_expanded,
        #             self.algebra.reversion(w_geom)
        #         )
        #     ),
        #     axis=-2
        # )
        # if self.bias is not None:
        #     b_geom = self.algebra.from_tensor(
        #         self.bias, self.blade_indices_bias)
        #     result += b_geom

        # return self.activation(result)

        inputs_expanded = inputs.unsqueeze(len(inputs.shape) - 2)
        result = self.algebra.geom_prod( w_geom, self.algebra.geom_prod(inputs_expanded, self.algebra.reversion(w_geom))).sum(dim=-2)
        if self.bias is not None:
            b_geom = self.algebra.from_tensor(self.bias, self.blade_indices_bias)
            result += b_geom
        if self.activation_fn:
            result = self.activation_fn(result)
        return result


# @register_keras_serializable(package="TFGA")
class GeometricProductElementwise(GeometricAlgebraLayer):
    """Performs the elementwise geometric product with a list of multivectors
    with as many elements as there are input units.

    Args:
        algebra: GeometricAlgebra instance to use for the parameters
        blade_indices_kernel: Blade indices to use for the kernel parameter
        blade_indices_bias: Blade indices to use for the bias parameter (if used)
    """

    def __init__(
        self,
        algebra: GeometricAlgebra,
        blade_indices_kernel: List[int],
        blade_indices_bias: Union[None, List[int]] = None,
        activation=None,
        use_bias=True,
        # kernel_initializer="glorot_uniform",
        # bias_initializer="zeros",
        # kernel_regularizer=None,
        # bias_regularizer=None,
        # activity_regularizer=None,
        # kernel_constraint=None,
        # bias_constraint=None,
        **kwargs
    ):
        # super().__init__(algebra=algebra, activity_regularizer=activity_regularizer, **kwargs)
        super().__init__(algebra=algebra, **kwargs)

        self.blade_indices_kernel = torch.tensor(blade_indices_kernel, dtype=torch.int64)
        if use_bias:
            self.blade_indices_bias = torch.tensor(blade_indices_bias, dtype=torch.int64)
        
        # self.blade_indices_kernel = blade_indices_kernel.to(dtype=torch.int64)
        # if use_bias:
        #     self.blade_indices_bias = blade_indices_bias.to(dtype=torch.int64)

        self.activation_fn = activations.get(activation)
        self.use_bias = use_bias
        # self.kernel_initializer = initializers.get(kernel_initializer)
        # self.bias_initializer = initializers.get(bias_initializer)
        # self.kernel_regularizer = regularizers.get(kernel_regularizer)
        # self.bias_regularizer = regularizers.get(bias_regularizer)
        # self.kernel_constraint = constraints.get(kernel_constraint)
        # self.bias_constraint = constraints.get(bias_constraint)
        self.built = False

    def build(self, input_shape: torch.Size):
        self.num_input_units = input_shape[-2]
        shape_kernel = [
            self.num_input_units,
            self.blade_indices_kernel.shape[0]
        ]
        self.kernel = nn.Parameter(1./np.prod(shape_kernel)*torch.randn(shape_kernel)).to(dtype=torch.float)
        if self.use_bias:
            shape_bias = [self.num_input_units,self.blade_indices_bias.shape[0]]
            self.bias = nn.Parameter(1./np.prod(shape_bias)*torch.randn(shape_bias)).to(dtype=torch.float)
        else:
            self.bias = None

        # self.kernel = self.add_weight(
        #     "kernel",
        #     shape=shape_kernel,
        #     initializer=self.kernel_initializer,
        #     regularizer=self.kernel_regularizer,
        #     constraint=self.kernel_constraint,
        #     dtype=self.dtype,
        #     trainable=True
        # )
        # if self.use_bias:
        #     shape_bias = [self.num_input_units,
        #                   self.blade_indices_bias.shape[0]]
        #     self.bias = self.add_weight(
        #         "bias",
        #         shape=shape_bias,
        #         initializer=self.bias_initializer,
        #         regularizer=self.bias_regularizer,
        #         constraint=self.bias_constraint,
        #         dtype=self.dtype,
        #         trainable=True
        #     )
        # else:
        #     self.bias = None
        self.built = True

    def compute_output_shape(self, input_shape):
        return torch.Size([*input_shape[:-1], self.algebra.num_blades])

    def forward(self, inputs):
        if not self.built: self.build(inputs.shape)
        w_geom = self.algebra.from_tensor(
            self.kernel, self.blade_indices_kernel)

        # Elementwise multiplication for each unit with a multivector.
        # [..., U, X] * [U, X] -> [..., U, X]
        result = self.algebra.geom_prod(inputs, w_geom)

        if self.bias is not None:
            b_geom = self.algebra.from_tensor(
                self.bias, self.blade_indices_bias)
            result += b_geom

        if self.activation_fn:
            result = self.activation_fn(result)
        return result

    def get_config(self):
        config = super().get_config()
        config.update({
            "blade_indices_kernel":
                self.blade_indices_kernel.cpu().detach().numpy(),
            "blade_indices_bias":
                self.blade_indices_bias.cpu().detach().numpy(),
            # "activation":
            #     self.activation,
                # activations.serialize(self.activation),
            "use_bias":
                self.use_bias,
            # "kernel_initializer":
            #     initializers.serialize(self.kernel_initializer),
            # "bias_initializer":
            #     initializers.serialize(self.bias_initializer),
            # "kernel_regularizer":
            #     regularizers.serialize(self.kernel_regularizer),
            # "bias_regularizer":
            #     regularizers.serialize(self.bias_regularizer),
            # "activity_regularizer":
            #     regularizers.serialize(self.activity_regularizer),
            # "kernel_constraint":
            #     constraints.serialize(self.kernel_constraint),
            # "bias_constraint":
            #     constraints.serialize(self.bias_constraint)
        })
        return config


# @register_keras_serializable(package="TFGA")
class GeometricSandwichProductElementwise(GeometricProductElementwise):
    """Performs the elementwise geometric sandwich product with a list of
    multivectors with as many elements as there are input units.

    Args:
        algebra: GeometricAlgebra instance to use for the parameters
        blade_indices_kernel: Blade indices to use for the kernel parameter
        blade_indices_bias: Blade indices to use for the bias parameter (if used)
    """

    def __init__(
        self, algebra, blade_indices_kernel, blade_indices_bias=None,
        activation=None, use_bias=True, 
        # kernel_initializer="glorot_uniform",
        # bias_initializer="zeros", kernel_regularizer=None,
        # bias_regularizer=None, activity_regularizer=None,
        # kernel_constraint=None, bias_constraint=None, 
        **kwargs
    ):
        super().__init__(
            algebra,
            blade_indices_kernel,
            blade_indices_bias=blade_indices_bias,
            activation=activation,
            use_bias=use_bias,
            # kernel_initializer=kernel_initializer,
            # bias_initializer=bias_initializer,
            # kernel_regularizer=kernel_regularizer,
            # bias_regularizer=bias_regularizer,
            # activity_regularizer=activity_regularizer,
            # kernel_constraint=kernel_constraint,
            # bias_constraint=bias_constraint, 
            **kwargs
        )

    def forward(self, inputs):
        if not self.built: self.build(inputs.shape)
        w_geom = self.algebra.from_tensor( self.kernel, self.blade_indices_kernel)

        # Elementwise multiplication Rx~R for each unit with a multivector.
        # [..., U, X] * [U, X] -> [..., U, X]
        result = self.algebra.geom_prod(
            w_geom,
            self.algebra.geom_prod(
                inputs,
                self.algebra.reversion(w_geom)
            )
        )

        if self.bias is not None:
            b_geom = self.algebra.from_tensor(
                self.bias, self.blade_indices_bias)
            result += b_geom

        if self.activation_fn:
            result = self.activation_fn(result)
        return result


# @register_keras_serializable(package="TFGA")
class GeometricProductConv1D(GeometricAlgebraLayer):
    """Analagous to Keras' Conv1D layer but using multivector-valued kernels
    instead of scalar ones and geometric product instead of
    standard multiplication.

    Args:
        algebra: GeometricAlgebra instance to use for the parameters
        filters: How many channels the output will have
        kernel_size: Size for the convolution kernel
        stride: Stride to use for the convolution
        padding: "SAME" (zero-pad input length so output
            length == input length / stride) or "VALID" (no padding)
        blade_indices_kernel: Blade indices to use for the kernel parameter
        blade_indices_bias: Blade indices to use for the bias parameter (if used)
    """

    def __init__(
        self,
        algebra: GeometricAlgebra,
        filters: int,
        kernel_size: int,
        stride: int,
        padding: str,
        blade_indices_kernel: List[int],
        blade_indices_bias: Union[None, List[int]] = None,
        dilations: Union[None, int] = None,
        activation=None,
        use_bias=True,
        # kernel_initializer="glorot_uniform",
        # bias_initializer="zeros",
        # kernel_regularizer=None,
        # bias_regularizer=None,
        # activity_regularizer=None,
        # kernel_constraint=None,
        # bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            algebra=algebra,
            # activity_regularizer=activity_regularizer,
            **kwargs
        )

        self.filters = filters
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilations = dilations

        self.blade_indices_kernel = torch.tensor( blade_indices_kernel, dtype=torch.int64)
        if use_bias:
            self.blade_indices_bias = torch.tensor( blade_indices_bias, dtype=torch.int64)
        # self.blade_indices_kernel = blade_indices_kernel.to(dtype=torch.int64)
        # if use_bias:
        #     self.blade_indices_bias = blade_indices_bias.to(dtype=torch.int64)

        self.activation_fn = activations.get(activation)
        self.use_bias = use_bias
        # self.kernel_initializer = initializers.get(kernel_initializer)
        # self.bias_initializer = initializers.get(bias_initializer)
        # self.kernel_regularizer = regularizers.get(kernel_regularizer)
        # self.bias_regularizer = regularizers.get(bias_regularizer)
        # self.kernel_constraint = constraints.get(kernel_constraint)
        # self.bias_constraint = constraints.get(bias_constraint)
        self.built = False

    def build(self, input_shape: torch.Size):
        # I: [..., S, C, B]
        self.num_input_filters = input_shape[-2]

        # K: [K, IC, OC, B]
        shape_kernel = [
            self.kernel_size,
            self.num_input_filters,
            self.filters,
            self.blade_indices_kernel.shape[0]
        ]
        self.kernel = nn.Parameter(1./np.prod(shape_kernel)*torch.randn(size=shape_kernel)).to(dtype=torch.float)
        if self.use_bias:
            shape_bias = [self.filters, self.blade_indices_bias.shape[0]]
            self.bias = nn.Parameter(1./np.prod(shape_bias)*torch.randn(size=shape_bias)).to(dtype=torch.float)
        else:
            self.bias = None

        # self.kernel = self.add_weight(
        #     "kernel",
        #     shape=shape_kernel,
        #     initializer=self.kernel_initializer,
        #     regularizer=self.kernel_regularizer,
        #     constraint=self.kernel_constraint,
        #     dtype=self.dtype,
        #     trainable=True
        # )
        # if self.use_bias:
        #     shape_bias = [self.filters, self.blade_indices_bias.shape[0]]
        #     self.bias = self.add_weight(
        #         "bias",
        #         shape=shape_bias,
        #         initializer=self.bias_initializer,
        #         regularizer=self.bias_regularizer,
        #         constraint=self.bias_constraint,
        #         dtype=self.dtype,
        #         trainable=True
        #     )
        # else:
        #     self.bias = None
        self.built = True

    def forward(self, inputs):
        if not self.built: 
            self.build(inputs.shape)
        k_geom = self.algebra.from_tensor(
            self.kernel, self.blade_indices_kernel)

        result = self.algebra.geom_conv1d(
            inputs, k_geom,
            stride=self.stride, padding=self.padding,
            dilations=self.dilations
        )

        if self.bias is not None:
            b_geom = self.algebra.from_tensor(
                self.bias, self.blade_indices_bias)
            result += b_geom

        if self.activation_fn:
            result = self.activation_fn(result)
        return result

    def get_config(self):
        config = super().get_config()
        config.update({
            "filters":
                self.filters,
            "kernel_size":
                self.kernel_size,
            "stride":
                self.stride,
            "padding":
                self.padding,
            "dilations":
                self.dilations,
            "blade_indices_kernel":
                self.blade_indices_kernel.numpy(),
            "blade_indices_bias":
                self.blade_indices_bias.numpy(),
            # "activation":
            #     activations.serialize(self.activation),
            "use_bias":
                self.use_bias,
            # "kernel_initializer":
            #     initializers.serialize(self.kernel_initializer),
            # "bias_initializer":
            #     initializers.serialize(self.bias_initializer),
            # "kernel_regularizer":
            #     regularizers.serialize(self.kernel_regularizer),
            # "bias_regularizer":
            #     regularizers.serialize(self.bias_regularizer),
            # "activity_regularizer":
            #     regularizers.serialize(self.activity_regularizer),
            # "kernel_constraint":
            #     constraints.serialize(self.kernel_constraint),
            # "bias_constraint":
            #     constraints.serialize(self.bias_constraint)

        })

        return config


# @register_keras_serializable(package="TFGA")
class GeometricAlgebraExp(GeometricAlgebraLayer):
    """Calculates the exponential function of the input. Input must square to
    a scalar.

    Args:
        algebra: GeometricAlgebra instance to use
        square_scalar_tolerance: Tolerance to use for the square scalar check
            or None if the check should be skipped
    """

    def __init__(
        self,
        algebra: GeometricAlgebra,
        square_scalar_tolerance: Union[float, None] = 1e-4,
        **kwargs
    ):
        super().__init__(algebra=algebra, **kwargs)
        self.square_scalar_tolerance = square_scalar_tolerance
        self.built = False

    def compute_output_shape(self, input_shape):
        return torch.Size([*input_shape[:-1], self.algebra.num_blades])

    def build(self,inputs_shape): self.built = True

    def forward(self, inputs):
        if not self.built: self.build(inputs.shape)
        return self.algebra.exp(
            inputs, square_scalar_tolerance=self.square_scalar_tolerance
        )

    def get_config(self):
        config = super().get_config()
        config.update({
            "square_scalar_tolerance": self.square_scalar_tolerance
        })
        return config
